from dexy.dexy_filter import DexyFilter
from dexy.filters.process_filters import SubprocessFilter
import os
import re
import shutil
import subprocess

class AsciidocFilter(SubprocessFilter):
    VERSION_COMMAND = "asciidoc --version"
    EXECUTABLE = "asciidoc"
    INPUT_EXTENSIONS = [".txt"]
    OUTPUT_EXTENSIONS = [".html", ".xml"]
    ALIASES = ['asciidoc']
    BINARY = False

    def command_string(self):
        if self.artifact.args.has_key('backend'):
            backend = self.artifact.args['backend']
            # TODO check file extension is valid for backend
        else:
            if self.artifact.ext == ".html":
                if self.version() >= "asciidoc 8.6.5":
                    backend = 'html5'
                else:
                    backend = 'html4'
            elif self.artifact.ext == ".xml":
                backend = "docbook45"
            elif self.artifact.ext == ".tex":
                backend = "latex"
            else:
                raise Exception("unexpected file extension in asciidoc filter %s" % self.artifact.ext)

        args = {
            'backend' : backend,
            'infile' : self.artifact.previous_artifact_filename,
            'outfile' : self.artifact.filename(),
            'prog' : self.executable(),
            'args' : self.command_line_args() or ""
        }

        return "%(prog)s -b %(backend)s %(args)s -o %(outfile)s %(infile)s" % args

class BibFilter(DexyFilter):
    INPUT_EXTENSIONS = [".tex"]
    OUTPUT_EXTENSIONS = [".tex"]
    ALIASES = ['bib']

    """Filter which replaces a hard-coded bibtex file with a .bib file specified as an input."""
    def process_text(self, input_text):
        for k, a in self.artifact.inputs().items():
            if a.filename().endswith("bib"):
                input_text = re.sub("bibliography{[^}]+}", "bibliography{%s}" % a.filename(), input_text)

        return input_text

class HtLatexFilter(DexyFilter):
    """
    Generates HTML from LaTeX source.
    """
    INPUT_EXTENSIONS = [".tex", ".txt"]
    OUTPUT_EXTENSIONS = [".html"]
    EXECUTABLES = ['htlatex']
    ALIASES = ['htlatex']
    FINAL = True

    def process(self):
        self.artifact.create_temp_dir()
        wf = os.path.join(self.artifact.temp_dir(), os.path.basename(self.artifact.name))

        f = open(wf, "w")
        f.write(self.artifact.input_text())
        f.close()

        if self.artifact.args.has_key('htlatex'):
            htlatex_args = self.artifact.args['htlatex']
        else:
            htlatex_args = ""


        command = "%s %s %s" % (self.__class__.executable(), os.path.basename(self.artifact.name), htlatex_args)
        self.log.info("running: %s" % command)
        env = None
        proc = subprocess.Popen(command, shell=True,
                                cwd=self.artifact.temp_dir(),
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                env=env)
        stdout, stderr = proc.communicate()
        self.artifact.stdout = stdout

        html_filename = wf.replace(".tex", ".html")

        # The main HTML file generated.
        f = open(html_filename, "r")
        self.artifact.set_data(f.read())
        f.close()

        for f in os.listdir(self.artifact.temp_dir()):
            basefilepath = os.path.dirname(self.artifact.name)
            f_ext = os.path.splitext(f)[1]

            # handle text-based output
            if (f_ext in [".css", ".html"]) and f != self.artifact.name.replace(".tex", ".html"):
                f_key = os.path.join(basefilepath, f)
                new_artifact = self.artifact.add_additional_artifact(f_key, f_ext)
                with open(os.path.join(self.artifact.temp_dir(), f), "r") as generated_file:
                    new_artifact.set_data(generated_file.read())

            # handle binary output
            if f_ext in [".png"]:
                f_key = os.path.join(basefilepath, f)
                new_artifact = self.artifact.add_additional_artifact(f_key, f_ext)
                shutil.copyfile(os.path.join(self.artifact.temp_dir(), f), new_artifact.filepath())

class LatexFilter(DexyFilter):
    """
    Generates a PDF file from LaTeX source.
    """
    INPUT_EXTENSIONS = [".tex", ".txt"]
    OUTPUT_EXTENSIONS = [".pdf", ".png"]
    EXECUTABLES = ['pdflatex', 'latex']
    ALIASES = ['latex']
    BINARY = True
    FINAL = True

    def process(self):
        latex_filename = self.artifact.filename().replace(self.artifact.ext, ".tex")
        self.artifact.generate_workfile(latex_filename)

        if self.artifact.args.has_key('env'):
            env = os.environ
            env.update(self.artifact.args['env'])
        else:
            env = None

        latex_command = "%s -interaction=batchmode %s" % (self.__class__.executable(), latex_filename)

        if self.__class__.executable_present("bibtex"):
            bibtex_command = "bibtex %s" % os.path.splitext(self.artifact.filename())[0]
        else:
            bibtex_command = None

        self.artifact.stdout = ""
        def run_cmd(command):
            self.log.info("running: %s" % command)
            proc = subprocess.Popen(command, shell=True,
                                    cwd=self.artifact.artifacts_dir,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    env=env)

            stdout, stderr = proc.communicate()
            self.artifact.stdout += stdout
            if proc.returncode > 2: # Set at 2 for now as this is highest I've hit, better to detect whether PDF has been generated?
                raise Exception("latex error, look for information in %s" %
                                latex_filename.replace(".tex", ".log"))
            elif proc.returncode > 0:
                self.log.warn("""A non-critical latex error has occurred running %s,
                status code returned was %s, look for information in %s""" % (
                self.artifact.key, proc.returncode,
                latex_filename.replace(".tex", ".log")))


        runbibtex = bibtex_command # TODO allow opting out of running bibtex in args
        if runbibtex:
            run_cmd(latex_command) #generate aux
            run_cmd(bibtex_command) #generate bbl
        run_cmd(latex_command) #first run
        run_cmd(latex_command) #second run - fix references

class EmbedFonts(SubprocessFilter):
    """
    Use to embed fonts and do other prepress as required for some types of printing.
    """
    INPUT_EXTENSIONS = [".pdf"]
    OUTPUT_EXTENSIONS = [".pdf"]
    EXECUTABLE = 'ps2pdf'
    ALIASES = ['embedfonts', 'prepress']

    def preprocess_command_string(self):
        pf = self.artifact.previous_artifact_filename
        af = self.artifact.filename()
        return "%s -dPDFSETTINGS=/prepress %s %s" % (self.EXECUTABLE, pf, af)

    def pdffonts_command_string(self):
        return "%s %s" % ("pdffonts", self.artifact.filename())

    def process(self):
        env = self.setup_env()

        proc, stdout = self.run_command(self.preprocess_command_string(), env)
        self.handle_subprocess_proc_return(proc.returncode, stdout)
        self.artifact.stdout = stdout

        proc, stdout = self.run_command(self.pdffonts_command_string(), env)
        self.handle_subprocess_proc_return(proc.returncode, stdout)
        self.artifact.stdout += stdout

class ROutputBatchFilter(SubprocessFilter):
    """Runs R code in batch mode. Uses the --slave flag so doesn't echo commands, just returns output."""
    ALIASES = ['rout', 'routbatch']
    EXECUTABLE = 'R CMD BATCH --vanilla --quiet --slave --no-timing'
    INPUT_EXTENSIONS = ['.txt', '.r', '.R']
    OUTPUT_EXTENSIONS = [".txt"]
    VERSION_COMMAND = "R --version"
    BINARY = False
    FINAL = False

class RBatchFilter(SubprocessFilter):
    """Runs R code in batch mode."""
    ALIASES = ['rintbatch']
    EXECUTABLE = 'R CMD BATCH --vanilla --quiet --no-timing'
    INPUT_EXTENSIONS = ['.txt', '.r', '.R']
    OUTPUT_EXTENSIONS = [".Rout", '.txt']
    VERSION_COMMAND = "R --version"
    BINARY = False
    FINAL = False

class Rd2PdfFilter(SubprocessFilter):
    INPUT_EXTENSIONS = [".Rd"]
    OUTPUT_EXTENSIONS = [".pdf", ".dvi"]
    EXECUTABLE = 'R CMD Rd2pdf'
    VERSION_COMMAND = 'R CMD Rd2pdf -v'
    ALIASES = ['rd2pdf', 'Rd2pdf']

    def command_string(self):
        title = os.path.splitext(self.artifact.name)[0].replace("_", " ")
        args = {
            'prog' : self.executable(),
            'out' : self.artifact.filename(),
            'in' : self.artifact.previous_artifact_filename,
            'title' : title
        }
        return "%(prog)s --output=%(out)s --title=\"%(title)s\" %(in)s" % args

class RagelRubySubprocessFilter(SubprocessFilter):
    """
    Generates ruby source code from a ragel file.
    """
    ALIASES = ['rlrb', 'ragelruby']
    BINARY = False
    EXECUTABLE = 'ragel -R'
    FINAL = False
    INPUT_EXTENSIONS = [".rl"]
    OUTPUT_EXTENSIONS = [".rb"]
    VERSION_COMMAND = 'ragel --version'

    def command_string(self):
        wf = self.artifact.previous_artifact_filename
        of = self.artifact.filename()
        return "%s %s -o %s" % (self.executable(), wf, of)

class Ps2PdfSubprocessFilter(SubprocessFilter):
    """
    Converts a postscript file to PDF format.
    """
    ALIASES = ['ps2pdf']
    EXECUTABLE = 'ps2pdf'
    INPUT_EXTENSIONS = [".ps", ".txt"]
    OUTPUT_EXTENSIONS = [".pdf"]

class Html2PdfSubprocessFilter(SubprocessFilter):
    """
    Renders HTML to PDF using wkhtmltopdf. If the HTML relies on assets such as
    CSS or image files, these should be specified as inputs.

    If you have an older version of wkhtmltopdf, and are running on a server,
    you may get XServer errors. You can install xvfb and run Dexy as
    "xvfb-run dexy". Or upgrade to the most recent wkhtmltopdf which only needs
    X11 client libs.
    """
    ALIASES = ['html2pdf', 'wkhtmltopdf']
    EXECUTABLE = 'wkhtmltopdf'
    INPUT_EXTENSIONS = [".html", ".txt"]
    OUTPUT_EXTENSIONS = [".pdf"]
    VERSION_COMMAND = 'wkhtmltopdf --version'

    def command_string(self):
        # Create a temporary directory and populate it with all inputs.
        self.artifact.create_temp_dir(populate=True)
        workfile = os.path.join(self.artifact.hashstring, self.artifact.previous_canonical_filename)

        args = {
            'prog' : self.executable(),
            'in' : workfile,
            'out' : self.artifact.filename()
        }
        return "%(prog)s %(in)s %(out)s" % args

class DotFilter(SubprocessFilter):
    """
    Renders .dot files to either PNG or PDF images.
    """
    INPUT_EXTENSIONS = [".dot"]
    OUTPUT_EXTENSIONS = [".png", ".pdf"]
    EXECUTABLE = 'dot'
    VERSION_COMMAND = 'dot -V'
    ALIASES = ['dot', 'graphviz']

    def command_string(self):
        args = {
            'prog' : self.executable(),
            'format' : self.artifact.ext.replace(".",""),
            'workfile' : self.artifact.previous_artifact_filename,
            'outfile' : self.artifact.filename()
        }
        return "%(prog)s -T%(format)s -o%(outfile)s %(workfile)s" % args

class Pdf2ImgSubprocessFilter(SubprocessFilter):
    """
    Converts a PDF file to a PNG image using ghostscript (subclass this to
    convert to other image types).

    Returns the image generated by page 1 of the PDF by default, the optional
    'page' parameter can be used to specify other pages.
    """
    ALIASES = ['pdf2img', 'pdf2png']
    EXECUTABLE = "gs"
    GS_DEVICE = 'png16m -r300'
    INPUT_EXTENSIONS = ['.pdf']
    OUTPUT_EXTENSIONS = ['.png']
    VERSION_COMMAND = "gs --version"

    def command_string(self):
        s = "%(prog)s -dSAFER -dNOPAUSE -dBATCH -sDEVICE=%(device)s -sOutputFile=%%d-%(out)s %(in)s"
        args = {
            'prog' : self.executable(),
            'device' : self.GS_DEVICE,
            'in' : self.artifact.previous_artifact_filename,
            'out' : self.artifact.filename()
        }
        return s % args

    def process(self):
        command = self.command_string()
        proc, stdout = self.run_command(command, self.setup_env())
        self.artifact.stdout = stdout
        self.handle_subprocess_proc_return(proc.returncode, stdout)

        if self.artifact.args.has_key('page'):
            page = self.artifact.args['page']
        else:
            page = 1

        page_file = os.path.join(self.artifact.artifacts_dir, "%s-%s" % (page, self.artifact.filename()))
        shutil.copyfile(page_file, self.artifact.filepath())

class Pdf2JpgSubprocessFilter(Pdf2ImgSubprocessFilter):
    ALIASES = ['pdf2jpg']
    GS_DEVICE = 'jpeg'
    OUTPUT_EXTENSIONS = ['.jpg']




from dexy.constants import Constants
from dexy.version import Version
from ordereddict import OrderedDict
import codecs
import dexy.introspect
import json
import glob
import hashlib
import inspect
import logging
import os
import shutil
import stat
import sys
import time
import traceback
import zlib

class Artifact(object):
    HASH_WHITELIST = Constants.ARTIFACT_HASH_WHITELIST
    MAX_DATA_DICT_DECIMALS = 5
    MAX_DATA_DICT_LENGTH = 10 ** MAX_DATA_DICT_DECIMALS
    META_ATTRS = [
        'additional',
        'binary_input',
        'binary_output',
        'created_by',
        'document_key',
        'ext',
        'final',
        'hashfunction',
        'initial',
        'is_last',
        'logstream',
        'key',
        'name',
        'output_hash',
        'state',
        'stdout'
    ]

    BINARY_EXTENSIONS = [
        '.gif',
        '.jpg',
        '.png',
        '.pdf',
        '.zip',
        '.tgz',
        '.gz',
        '.eot',
        '.ttf',
        '.woff',
        '.sqlite',
        '.sqlite3',
        '.swf'
    ]

    def __init__(self):
        if not hasattr(self.__class__, 'FILTERS'):
            self.__class__.FILTERS = dexy.introspect.filters(Constants.NULL_LOGGER)

        self._inputs = {}
        self.additional = None
        self.args = {}
        self.args['globals'] = {}
        self.artifacts_dir = 'artifacts' # TODO don't hard code
        self.batch_id = None
        self.batch_order = None
        self.binary_input = None
        self.binary_output = None
        self.controller_args = {}
        self.controller_args['globals'] = {}
        self.created_by = None
        self.ctime = None
        self.data_dict = OrderedDict()
        self.dexy_version = Version.VERSION
        self.dirty = False
        self.document_key = None
        self.elapsed = 0
        self.ext = None
        self.final = None
        self.finish_time = None
        self.hashfunction = 'md5'
        self.initial = None
        self.inode = None
        self.input_data_dict = OrderedDict()
        self.is_last = False
        self.key = None
        self.log = logging.getLogger()
        self.logstream = ""
        self.mtime = None
        self.name = None
        self.source = None
        self.start_time = None
        self.state = 'new'
        self.stdout = None

    def is_complete(self):
        return str(self.state) == 'complete'

    @classmethod
    def retrieve(klass, hashstring, hashfunction='md5'):
        if not hasattr(klass, 'retrieved_artifacts'):
            klass.retrieved_artifacts = {}
        if klass.retrieved_artifacts.has_key(hashstring):
            return klass.retrieved_artifacts[hashstring]
        else:
            artifact = klass()
            artifact.hashstring = hashstring
            artifact.hashfunction = hashfunction
            artifact.load()
            klass.retrieved_artifacts[hashstring] = artifact
            return artifact

    def load(self):
        self.load_meta()
        self.load_input()
        if self.is_complete() and not self.is_loaded():
            self.load_output()

    def load_inputs(self):
        for a in self.inputs():
            a.load()

    def save(self):
        if self.is_abstract():
            pass # For testing.
        elif not self.hashstring:
            raise Exception("can't persist an artifact without a hashstring!")
        else:
            self.save_meta()
            if self.is_complete() and not self.is_output_cached():
                try:
                    self.save_output()
                except IOError as e:
                    print "An error occured while saving %s" % self.key
                    raise e

    def is_abstract(self):
        return not hasattr(self, 'save_meta')

    def setup_initial(self):
        """
        Set up an initial artifact (the first artifact in a document's filter chain).
        """
        self._inputs = self.doc.input_artifacts()
        self.binary_input = (self.doc.ext in self.BINARY_EXTENSIONS)
        self.binary_output = self.binary_input
        self.ext = self.doc.ext
        self.initial = True

        if self.args.has_key('final'):
            self.final = self.args['final']
        elif os.path.basename(self.name).startswith("_"):
            self.final = False

        if not self.doc.virtual:
            stat_info = os.stat(self.name)
            self.ctime = stat_info[stat.ST_CTIME]
            self.mtime = stat_info[stat.ST_MTIME]
            self.inode = stat_info[stat.ST_INO]

        self.set_data(self.doc.initial_artifact_data())

        # TODO remove?
        if not self.data_dict:
            raise Exception("no data dict!")
        elif len(self.data_dict) == 0:
            raise Exception("data dict has len 0!")

        self.state = 'complete'

    def setup_from_filter_class(self):
        # cache filter class source code so it only has to be calculated once
        if not hasattr(self.filter_class, 'SOURCE_CODE'):
            # get source code of this filter class + all parent filter classes.
            source = ""
            klass = self.filter_class

            # get source code from filter class and all parent classes
            while klass != dexy.dexy_filter.DexyFilter:
                source += inspect.getsource(klass)
                klass = klass.__base__

            # and then get source code of DexyFilter class
            source += inspect.getsource(dexy.dexy_filter.DexyFilter)

            filter_class_source = source
            self.filter_class.SOURCE_CODE = self.compute_hash(filter_class_source)

        if not hasattr(self.filter_class, 'VERSION'):
            filter_version = self.filter_class.version(self.log)
            self.filter_class.VERSION = filter_version

        self.filter_name = self.filter_class.__name__
        self.filter_source = self.filter_class.SOURCE_CODE
        self.filter_version = self.filter_class.VERSION

        if self.final is None:
            self.final = self.filter_class.FINAL

    def setup_from_previous_artifact(self, previous_artifact):
        for a in ['final', 'mtime', 'ctime', 'inode']:
                setattr(self, a, getattr(previous_artifact, a))

        self._inputs.update(previous_artifact.inputs())
        # Need to loop over each artifact's inputs in case extra ones have been
        # added anywhere.
        for k, a in previous_artifact.inputs().iteritems():
            self._inputs.update(a.inputs())
        self.binary_input = previous_artifact.binary_output
        self.input_data_dict = previous_artifact.data_dict
        self.input_ext = previous_artifact.ext
        self.previous_artifact_filename = previous_artifact.filename()
        self.previous_artifact_filepath = previous_artifact.filepath()
        self.previous_canonical_filename = previous_artifact.canonical_filename(True)

        # The JSON output of previous artifact
        if not previous_artifact.binary_output:
            self.previous_cached_output_filepath = previous_artifact.cached_output_filepath()

        # Determine file extension of output
        if hasattr(self, 'next_filter_class'):
            next_inputs = self.next_filter_class.INPUT_EXTENSIONS
        else:
            next_inputs = None

        self.ext = self.filter_class.output_file_extension(
                previous_artifact.ext,
                self.name,
                next_inputs)

        self.binary_output = self.filter_class.BINARY
        if self.binary_output is None:
            self.set_binary_from_ext()

        self.state = 'setup'

    @classmethod
    def setup(klass, doc, artifact_key, filter_class = None, previous_artifact = None):
        """
        Create an Artifact instance and load all information needed to
        calculate its hashstring.
        """
        artifact = klass()
        artifact.key = artifact_key
        artifact.filter_class = filter_class

        # Add references for convenience
        artifact.artifacts_dir = doc.artifacts_dir
        artifact.controller_args = doc.controller.args
        artifact.hashfunction = doc.controller.args['hashfunction']
        artifact.db = doc.db
        artifact.doc = doc
        artifact.log = doc.log

        # These attributes are the same for all artifacts pertaining to a document
        artifact.args = doc.args
        artifact.batch_id = doc.batch_id
        artifact.document_key = doc.key()
        artifact.name = doc.name

        # Set batch order to next in sequence
        artifact.batch_order = artifact.db.next_batch_order(artifact.batch_id)

        next_filter_class = doc.next_filter_class()
        if next_filter_class:
            artifact.next_filter_name = next_filter_class.__name__
            artifact.next_filter_class = next_filter_class

        if previous_artifact:
            artifact.setup_from_previous_artifact(previous_artifact)
            artifact.setup_from_filter_class()
        else:
            artifact.setup_initial()

        artifact.set_hashstring()
        return artifact

    def run(self):
        start = time.time()

        if self.controller_args['nocache'] or not self.is_complete():
            # We have to actually run things...
            if not self.filter_class:
                self.filter_class = dexy.introspect.get_filter_by_name(self.filter_name, self.doc.__class__.filter_list)

            # Set up instance of filter.
            filter_instance = self.filter_class()
            filter_instance.artifact = self
            filter_instance.log = self.log

            try:
                filter_instance.process()
            except Exception as e:
                print "Error occurred while running", self.key
                x, y, tb = sys.exc_info()
                print "Original traceback:"
                traceback.print_tb(tb)
                pattern = os.path.join(self.artifacts_dir, self.hashstring)
                files_matching = glob.glob(pattern)
                if len(files_matching) > 0:
                    print "Here are working files which might have clues about this error:"
                    for f in files_matching:
                        print f
                raise e

            h = hashlib.sha512()

            if self.data_dict and len(self.data_dict) > 0:
                h.update(self.output_text().encode("utf-8"))

            elif self.is_canonical_output_cached:
                self.state = 'complete'
                self.save()

                f = open(self.filepath(), "rb")
                while True:
                    data = f.read(h.block_size)
                    if not data:
                        break
                    h.update(data)

            else:
                raise Exception("data neither in memory nor on disk")

            self.output_hash = h.hexdigest()
            self.logstream = self.doc.logstream.getvalue()
            self.state = 'complete'
            self.source = 'run'
            self.save()
        else:
            self.source = 'cache'
            self.log.debug("using cached artifact for %s" % self.key)

            # make sure additional artifacts are added to db
            for a in self.inputs().values():
                if a.created_by == self.key:
                    if not a.additional:
                        raise Exception("created_by should only apply to additional artifacts")
                    # TODO Should this be done in Artifact.retrieve?
                    a.batch_id = self.batch_id
                    self.db.append_artifact(a)

        self.elapsed = time.time() - start
        self.db.update_artifact(self)

    def add_additional_artifact(self, key_with_ext, ext):
        """create an 'additional' artifact with random hashstring"""
        new_artifact = self.__class__()
        new_artifact.key = key_with_ext
        if ext.startswith("."):
            new_artifact.ext = ext
        else:
            new_artifact.ext = ".%s" % ext
        new_artifact.final = True
        new_artifact.hashfunction = self.hashfunction
        new_artifact.additional = True
        new_artifact.set_binary_from_ext()
        new_artifact.artifacts_dir = self.artifacts_dir
        new_artifact.inode = self.hashstring
        new_artifact.created_by = self.document_key
        # TODO filter class source?

        # TODO this is duplicated in setup_from_previous_artifact, should reorganize
        for at in ['batch_id', 'document_key', 'mtime', 'ctime']:
                val = getattr(self, at)
                setattr(new_artifact, at, val)

        new_artifact.set_hashstring()
        self.add_input(key_with_ext, new_artifact)
        self.db.append_artifact(new_artifact) # append to db because not part of doc.artifacts
        return new_artifact

    def add_input(self, key, artifact):
        self._inputs[key] = artifact

    def inputs(self):
        return self._inputs

    def set_binary_from_ext(self):
        # TODO list more binary extensions or find better way to do this
        if self.ext in self.BINARY_EXTENSIONS:
            self.binary_output = True
        else:
            self.binary_output = False

    def set_data(self, data):
        self.data_dict['1'] = data

    def set_data_from_artifact(self):
        f = codecs.open(self.filepath(), "r", encoding="utf-8")
        self.data_dict['1'] = f.read()

    def is_loaded(self):
        return hasattr(self, 'data_dict') and len(self.data_dict) > 0

    def compute_hash(self, text):
        unicode_text = None

        if type(text) == unicode:
            unicode_text = text
        elif type(text) in [dict, list]:
            unicode_text = json.dumps(text)
        elif self.binary_input:
            pass
        else:
            unicode_text = unicode(text, encoding="utf-8")

        if unicode_text:
            text = unicode_text.encode("utf-8")

        if self.hashfunction == 'md5':
            h = hashlib.md5(text).hexdigest()

        elif self.hashfunction == 'sha1':
            h = hashlib.sha1(text).hexdigest()

        elif self.hashfunction == 'sha224':
            h = hashlib.sha224(text).hexdigest()

        elif self.hashfunction == 'sha256':
            h = hashlib.sha256(text).hexdigest()

        elif self.hashfunction == 'sha384':
            h = hashlib.sha384(text).hexdigest()

        elif self.hashfunction == 'sha512':
            h = hashlib.sha512(text).hexdigest()

        elif self.hashfunction == 'crc32':
            h = str(zlib.crc32(text) & 0xffffffff)

        elif self.hashfunction == 'adler32':
            h = str(zlib.adler32(text) & 0xffffffff)

        else:
            raise Exception("unexpected hash function %s" % self.hashfunction)

        return h

    def input_hashes(self):
        """
        Returns an OrderedDict of key, hashstring for each input artifact, sorted by key.
        """
        return OrderedDict((k, str(self.inputs()[k].hashstring)) for k in sorted(self.inputs()))

    def hash_dict(self):
        """
        Calculate and cache the elements used to compute the hashstring
        """
        if not hasattr(self.__class__, 'SOURCE_CODE'):
            artifact_class_source = inspect.getsource(self.__class__)
            artifact_py_source = inspect.getsource(Artifact)
            self.__class__.SOURCE_CODE = self.compute_hash(artifact_class_source + artifact_py_source)

        self.artifact_class_source = self.__class__.SOURCE_CODE

        if self.dirty:
            self.dirty_string = time.gmtime()

        hash_dict = OrderedDict()

        hash_dict['inputs'] = self.input_hashes()

        for k in self.HASH_WHITELIST:
            if self.__dict__.has_key(k):
                v = self.__dict__[k]
                if hasattr(v, 'items'):
                    hash_v = OrderedDict()
                    for k1 in sorted(v.keys()):
                        v1 = v[k1]
                        try:
                            if len(str(v1)) > 50:
                                raise Exception()
                            json.dumps(v1)
                            hash_v[str(k1)] = v1
                        except Exception:
                            # Use a hash if we will have problems saving to JSON
                            # or if the data is large (don't want to clutter up the DB,
                            # makes it harder to spot differences)
                            hash_v[str(k1)] = self.compute_hash(v1)
                else:
                    hash_v = str(v)
                hash_dict[str(k)] = hash_v
        return hash_dict

    def set_hashstring(self):
        if hasattr(self, 'hashstring'):
            raise Exception("setting hashstring twice")

        hash_data = str(self.hash_dict())
        self.hashstring = self.compute_hash(hash_data)

        try:
            original_document_key = self.document_key
            if not self.is_loaded():
                self.load()
            self.document_key = original_document_key
        except AttributeError as e:
            if not self.is_abstract():
                raise e
        except IOError as e:
            self.save_meta()


    def convert_if_not_unicode(self, s):
        if type(s) == unicode:
            return s
        else:
            return unicode(s, encoding="utf-8")

    def input_text(self):
        return u"".join([self.convert_if_not_unicode(v) for k, v in self.input_data_dict.items()])

    def output_text(self):
        return u"".join([self.convert_if_not_unicode(v) for k, v in self.data_dict.items()])

    def relative_refs(self, relative_to_file):
        """How to refer to this artifact, relative to another."""

        doc_dir = os.path.dirname(relative_to_file)
        return [
                os.path.relpath(self.key, doc_dir),
                os.path.relpath(self.long_canonical_filename(), doc_dir),
                "/%s" % self.key,
                "/%s" % self.long_canonical_filename()
        ]

    def use_canonical_filename(self):
        """Returns the canonical filename after saving contents under this name
        in the artifacts directory."""
        self.write_to_file(os.path.join(self.artifacts_dir,
                                        self.canonical_filename()))
        return self.canonical_filename()

    def write_to_file(self, filename):
        dirname = os.path.dirname(filename)
        if not os.path.exists(dirname) and not dirname == '':
            os.makedirs(dirname)
        shutil.copyfile(self.filepath(), filename)

    def work_filename(self):
        return "%s.work%s" % (self.hashstring, self.input_ext)

    def generate_workfile(self, work_filename = None):
        if not work_filename:
            work_filename = self.work_filename()
        work_path = os.path.join(self.artifacts_dir, work_filename)
        work_file = codecs.open(work_path, "w", encoding="utf-8")
        work_file.write(self.input_text())
        work_file.close()

    def temp_filename(self, ext):
        return "%s.work%s" % (self.hashstring, ext)

    def open_tempfile(self, ext):
        tempfile_path = os.path.join(self.artifacts_dir, self.temp_filename(ext))
        codecs.open(tempfile_path, "w", encoding="utf-8")

    def temp_dir(self):
        return os.path.join(self.artifacts_dir, self.hashstring)

    def create_temp_dir(self, populate=False):
        tempdir = self.temp_dir()
        shutil.rmtree(tempdir, ignore_errors=True)
        os.mkdir(tempdir)

        if populate:
            # write all inputs to this directory, under their canonical names
            for input_artifact in self._inputs.values():
                filename = os.path.join(tempdir, input_artifact.canonical_filename())
                if os.path.exists(input_artifact.filepath()):
                    input_artifact.write_to_file(filename)
                    self.log.debug("Populating temp dir for %s with %s" % (self.key, filename))
                else:
                    self.log.warn("Skipping file %s for temp dir for %s, file does not exist (yet)" % (filename, self.key))

            # write the workfile to this directory under its canonical name
            previous = self.previous_artifact_filepath
            workfile = os.path.join(tempdir, self.previous_canonical_filename)
            if not os.path.exists(os.path.dirname(workfile)):
                os.makedirs(os.path.dirname(workfile))
            shutil.copyfile(previous, workfile)

    def canonical_dir(self, ignore_args = False):
        return os.path.dirname(self.name)

    def canonical_basename(self, ignore_args = False):
        return os.path.basename(self.canonical_filename(ignore_args))

    def canonical_filename(self, ignore_args = False):
        fn = os.path.splitext(self.key.split("|")[0])[0]

        if self.args.has_key('canonical-name') and not ignore_args:
            parent_dir = os.path.dirname(fn)
            return os.path.join(parent_dir, self.args['canonical-name'])
        elif self.args.has_key('postfix') and not ignore_args:
            return "%s%s%s" % (fn, self.ext, self.args['postfix'])
        else:
            return "%s%s" % (fn, self.ext)

    def long_canonical_filename(self):
        return "%s%s" % (self.key.replace("|", "-"), self.ext)

    def filename(self):
        """
        The filename where artifact content is stored, based on the hashstring.
        """
        if not hasattr(self, 'ext'):
            raise Exception("artifact %s has no ext" % self.key)
        return "%s%s" % (self.hashstring, self.ext)

    def filepath(self):
        """
        Full path (including artifacts dir location) to location where artifact content is stored.
        """
        return os.path.join(self.artifacts_dir, self.filename())

    def abs_filepath(self):
        return os.path.abspath(self.filepath())

    def breadcrumbs(self):
        """A list of parent dirs, plus the filename if it's not 'index.html'."""
        parent_dirs = os.path.dirname(self.canonical_filename()).split("/")

        if self.canonical_basename() == "index.html":
            result = parent_dirs
        else:
            result = parent_dirs.append(self.canonical_basename())

        if not result:
            result = []

        return result

    def titleized_name(self):
        if self.canonical_basename() == "index.html":
            return self.breadcrumbs()[-1].replace("-"," ").title()
        else:
            return os.path.splitext(self.canonical_basename())[0].replace("-"," ").title()

    def unique_key(self):
        return "%s:%s:%s" % (self.batch_id, self.document_key, self.key)

    def web_safe_document_key(self):
        # TODO this might not be unique
        return self.document_key.replace("/","-").replace("|", "-")

    def url(self):
        # TODO test for final
        return "/%s" % self.canonical_filename()

    def hyperlink(self, link_text = None):
        # TODO test for final
        if not link_text:
            link_text = self.canonical_basename()

        return """<a href="%s">%s</a>""" % (self.url(), link_text)

    def iframe(self, link_text = None, width = "600px", height = "300px"):
        # TODO test for final
        args = {
                'url' : self.url(),
                'hyperlink' : self.hyperlink(link_text),
                'width' : width,
                'height' : height
        }

        return """
<iframe src="%(url)s" width="%(width)s" height="%(height)s" style="border: thin solid gray;">
%(hyperlink)s
</iframe>
        """ % args

    def img(self):
        # TODO test for final
        return """<img src="/%s" alt="Image generated by dexy %s" />""" % (self.canonical_filename(), self.key)

    def has_sections(self):
        return (self.data_dict.keys() != ['1'])

    def relative_path_to_input(self, input_artifact):
        my_dir = os.path.dirname(self.name)
        input_dir = os.path.dirname(input_artifact.name)
        if my_dir == input_dir:
            relpath = ""
        else:
            relpath = os.path.relpath(input_dir, my_dir)
        return relpath

    def relative_key_for_input(self, input_artifact):
        relpath = self.relative_path_to_input(input_artifact)
        return os.path.join(relpath, os.path.basename(input_artifact.key))

    def convert_numbered_dict_to_ordered_dict(self, numbered_dict):
        ordered_dict = OrderedDict()
        for x in sorted(numbered_dict.keys()):
            k = x.split(":", 1)[1]
            ordered_dict[k] = numbered_dict[x]
        return ordered_dict

    def convert_data_dict_to_numbered_dict(self):
        if len(self.data_dict) >= self.MAX_DATA_DICT_LENGTH:
            exception_msg = """Your data dict has %s items, which is greater than the arbitrary limit of %s items.
            You can increase this limit by changing MAX_DATA_DICT_DECIMALS."""
            raise Exception(exception_msg % (len(self.data_dict), self.MAX_DATA_DICT_LENGTH))

        data_dict = {}
        i = -1
        for k, v in self.data_dict.iteritems():
            i += 1
            fmt = "%%0%sd:%%s" % self.MAX_DATA_DICT_DECIMALS
            data_dict[fmt % (i, k)] = v
        return data_dict



from dexy.dexy_filter import DexyFilter
import json
import os
import re

class BlogFilter(DexyFilter):
    ALIASES = ['blogfilter']
    BLOG_CONFIG_FILE = 'blog-config.json'
    PASSWORD_KEYS = ['password', 'pass', 'pw']
    USERNAME_KEYS = ['username', 'user']

    def load_blog_conf(self):
        if not os.path.exists(self.BLOG_CONFIG_FILE):
            raise Exception("Could not find config file called %s" % self.BLOG_CONFIG_FILE)
        f = open(self.BLOG_CONFIG_FILE, "r")
        self.blog_conf = json.load(f)
        f.close()

        self.password = None
        for k in self.PASSWORD_KEYS:
            if self.blog_conf.has_key(k):
                self.password = self.blog_conf[k]
        if not self.password:
            raise Exception("none of %s was found in blog conf file %s, need to set a password" % (",".join(self.PASSWORD_KEYS), self.BLOG_CONFIG_FILE))

        self.username = None
        for k in self.USERNAME_KEYS:
            if self.blog_conf.has_key(k):
                self.username = self.blog_conf[k]
        if not self.username:
            raise Exception("none of %s was found in blog conf file %s, need to set a username" % (",".join(self.USERNAME_KEYS), self.BLOG_CONFIG_FILE))


    def load_post_conf(self):
        matches = [k for k in self.artifact.inputs().keys() if k.endswith("post.json|dexy")]
        if len(matches) == 0:
            raise Exception("no input found matching post.json|dexy! Make sure you create a post.json file, run it through the dexy filter, and make it an input to your blog post")
        self.k = matches[0]
        self.post_conf = json.loads(self.artifact.inputs()[self.k].output_text())

    def process_text(self, input_text):
        self.load_blog_conf()
        self.load_post_conf()
        api = self.initialize_api()

        if self.post_conf.has_key('post-id'):
            post_id = self.post_conf['post-id']
            self.update_post(api, input_text, self.post_conf['post-id'])
        else:
            post_id = self.new_post(api, input_text)
            self.post_conf['post-id'] = post_id

            json_file = re.sub('\|dexy$', "", self.k)
            f = open(json_file, 'w')
            json.dump(self.post_conf, f)
            f.close()

        return "%s" % post_id



from dexy.constants import Constants
import StringIO
import codecs
import dexy.controller
import dexy.introspect
import fnmatch
import hashlib
import json
import logging
import os
import re
import shutil
import time
import urllib2

try:
    import git
    USE_GIT = True
except ImportError:
    USE_GIT = False

class Document(object):

    def __init__(self):
        # initialize attributes
        self.args = {}
        self.artifacts = []
        self.elapsed = 0
        self.input_args = []
        self.input_keys = []
        self.inputs = []
        self.log = Constants.NULL_LOGGER # proper logging set up after we know our name
        self.logstream = StringIO.StringIO()
        self.priority = 10
        self.timing = []
        self.use_all_inputs = False

    def document_info(self):
        return {
            "artifacts" : [[a.hashstring, a.source, a.elapsed] for a in self.artifacts],
            "log" : self.logstream.getvalue(),
            "args" : self.args,
            "inputs" : [doc.key() for doc in self.inputs],
            "elapsed" : self.elapsed,
            "timing" : self.timing
        }

    def set_controller(self, controller):
        self.controller = controller

        # Set convenient attrs for info we use a lot.
        self.artifacts_dir = controller.args['artifactsdir']
        self.logsdir = controller.args['logsdir']
        self.artifact_class = controller.artifact_class
        self.db = controller.db
        self.batch_id = controller.batch_id

    def setup_log(self):
        self.log = logging.getLogger(self.key())
        self.log.setLevel(logging.DEBUG) # TODO Allow this to be changed.
        self.log.propagate = 0 # Stops logs being written to STDOUT if another library redirects root logger.
        handler = logging.StreamHandler(self.logstream)
        formatter = logging.Formatter(Constants.DEFAULT_LOGFORMAT)
        handler.setFormatter(formatter)
        self.log.addHandler(handler)

        try:
            self.log.addHandler(logging.getLogger('dexy').handlers[0])
        except IndexError:
            pass

    def set_name_and_filters(self, name_or_key, filters=[]):
        self.name = name_or_key.split("|")[0]
        self.ext = os.path.splitext(self.name)[1]
        self.filters = name_or_key.split("|")[1:]
        self.filters += filters

    def get_filter_for_alias(self, alias):
        """
        Given a filter alias, return the corresponding filter class.
        """
        return dexy.introspect.get_filter_for_alias(alias, self.__class__.filter_list)

    def final_artifact(self):
        return self.last_artifact

    def key(self):
        return "|".join([self.name] + self.filters)

    def add_input_key(self, input_key):
        if not input_key in self.input_keys:
            self.input_keys.append(input_key)

    def finalize_inputs(self, members_dict):
        """
        Called during setup, this method resolves which other docs this doc
        will depend on (have as inputs).
        """
        start = time.time()
        if len(self.input_args) > 0 or len(self.input_keys) > 0 or self.use_all_inputs:
            for doc in members_dict.values():
                if doc.key() in self.input_keys:
                    self.inputs.append(doc)
                else:
                    # Work out relative paths
                    dir1 = os.path.dirname(doc.name)
                    dir2 = os.path.dirname(self.name)
                    if dir1 == '':
                        dir1 = "."
                    if dir2 == "":
                        dir2 = "."

                    relpath = os.path.relpath(dir1, dir2)
                    relpath2 = os.path.relpath(dir2, dir1)

                    # Whether the document is in parent dir of input
                    in_parent_dir = not (".." in relpath) and not (relpath == ".")
                    in_same_dir = (relpath == ".") and (relpath2 == ".")
                    in_parent_or_child = (relpath == ".") or ((not ".." in relpath) ^ (not ".." in relpath2))

                    # See if input matches an item in inputs
                    specified = False
                    for input_glob in self.input_args:
                        # The full doc key is specified
                        is_exact_absolute_match = (input_glob == doc.key())

                        # The relative path doc key is specified
                        is_exact_relative_match = (os.path.join(relpath2, input_glob) == doc.key())

                        is_exact_match_in_parent_or_child_dir = (in_parent_or_child and input_glob == os.path.basename(doc.key()))

                        # A glob matches in any child dir
                        is_glob_match_in_child_dir = (in_parent_dir or in_same_dir) and fnmatch.fnmatch(os.path.basename(doc.key()), input_glob)

                        specified = (is_exact_absolute_match or is_exact_relative_match or is_exact_match_in_parent_or_child_dir or is_glob_match_in_child_dir)

                        if specified:
                           break

                    if self.args.has_key('exact-inputs'):
                        for exact_input in self.args['exact-inputs']:
                            if not specified:
                                specified = exact_input == doc.key()

                    # Work out relative priority
                    higher_priority = (self.priority > doc.priority)
                    equal_priority = (self.priority == doc.priority)

                    use_all_inputs = self.use_all_inputs
                    also_all_inputs = doc.use_all_inputs

                    all_inputs_and_qualified = use_all_inputs and (higher_priority or (equal_priority and in_parent_dir) or not also_all_inputs)
                    disqualified_by_strictinherit = self.controller.args['strictinherit'] and not (in_parent_or_child)

                    if specified or (all_inputs_and_qualified and not disqualified_by_strictinherit):
                        self.inputs.append(doc)

        elapsed = time.time() - start
        self.timing.append(("finalize-inputs", elapsed))

    def next_filter_alias(self):
        if self.at_last_step():
            return None
        else:
            return self.filters[self.step]

    def next_filter_class(self):
        alias = self.next_filter_alias()
        if alias:
            return self.get_filter_for_alias(alias)

    def at_last_step(self):
        return (len(self.filters) == self.step)

    def input_artifacts(self):
        input_artifacts = {}
        for i in self.inputs:
            k = i.key()
            a = i.final_artifact()
            if not a:
                raise Exception("No final artifact for document %s" % i.key())
            input_artifacts[k] = a
        return input_artifacts

    def initial_artifact_data(self):
        # put *safe* methods at top of list
        # check for safe methods first in controller
        # all other methods assumed to be dangerous and require 'danger' flag
        if self.args.has_key('contents'):
            data = self.args['contents']

        # potentially dangerous methods ...
        elif self.args.has_key('url'):
            url = self.args['url']

            local_filename = re.sub('[^\w\s-]', '', url).strip().lower()
            local_filename = re.sub('[-\s]+', '-', local_filename)
            filename = os.path.join(self.artifacts_dir, local_filename)
            header_filename = "%s.headers" % filename

            if not os.path.exists(os.path.dirname(filename)):
                os.makedirs(os.path.dirname(filename))

            header_dict = {}
            if os.path.exists(header_filename):
                header_file = codecs.open(header_filename, "r", encoding="utf-8")
                header_dict = json.load(header_file)
                header_file.close()

            request = urllib2.Request(url)

            # TODO add an md5 of the file to the header dict so we can check
            # that the etag/last-modified header is the corresponding one
            # TODO invalidate the hash if URL has changed
            # TODO get modification time of file via ftplib
            # TODO implement handlers for version control systems
            # TODO use FileHandler for default case?

            # Add any custom headers...
            # note that headers are ignored by anything other than http
            if header_dict.has_key('ETag') and os.path.exists(filename):
                request.add_header('If-None-Match', header_dict['ETag'])
            elif header_dict.has_key('Last-Modified') and os.path.exists(filename):
                request.add_header('If-Modifed-Since', header_dict['Last-Modified'])

            if self.controller.args['uselocals'] and os.path.exists(filename):
                f = codecs.open(filename, "r", encoding="utf-8")
                data = f.read()
                f.close()
            else:
                if self.controller.args['uselocals']:
                    print "local file %s not found, fetching remote url" % filename

                try:
                    u = urllib2.urlopen(request)
                    print "downloading contents of %s" % url
                    url_contents = u.read()

                    # Save the contents in our local cache
                    f = open(filename, "wb")
                    f.write(url_contents)
                    f.close()

                    # Save header info in our local cache
                    header_dict = {}
                    for s in u.info().headers:
                        a = s.partition(":")
                        header_dict[a[0]] = a[2].strip()
                    json.dump(header_dict, codecs.open(header_filename, "w", encoding="utf-8"))

                    data = url_contents
                except (urllib2.HTTPError, urllib2.URLError) as err:
                    if not hasattr(err, 'code'):
                        print "Trying to fetch %s for %s" % (url, self.name)
                        raise err

                    elif err.code == 304:
                        if os.path.exists(filename):
                            print "received NOT MODIFIED (304) from server, so using contents of %s" % (filename)
                            f = codecs.open(filename, "r", encoding="utf-8")
                            data = f.read()
                            f.close()
                        else:
                            raise err

                    elif err.code == 404:
                        raise Exception("url %s NOT FOUND (404) (for virtual file %s)" % (url, self.name))

                    else:
                        print "Trying to fetch %s for %s" % (url, self.name)
                        raise err

        elif self.args.has_key('repo') and self.args.has_key('path'):
            if not USE_GIT:
                raise Exception("you can't use repo/path unless you install GitPython")
            repo_url = self.args['repo']
            digest = hashlib.md5(repo_url).hexdigest()
            local_repo_dir = os.path.join(self.artifacts_dir, "repository-%s" % digest)

            if os.path.exists(local_repo_dir):
                repo = git.Repo(local_repo_dir)
                o = repo.remotes.origin
                assert o.url == repo_url, "local repo exists but url %s does not match requested url %s" % (o.url, repo_url)
                o.pull() # TODO remember last pulled time so that don't do this too often? what if network unavailable?
            else:
                repo = git.Repo.clone_from(repo_url, local_repo_dir)

            if self.args.has_key('commit'):
                commit = self.args['commit']
                tree = repo.commit(commit).tree
            else:
                tree = repo.heads.master.commit.tree

            try:
                blob = tree[self.args['path']]
                data = blob.data_stream.read()
            except KeyError as e:
                print "error processing virtual file", self.key(), "can't find path", self.args['path'], "in", repo_url
                raise e

        elif self.virtual:
            data = None

        else:
            # Normal local file, just read the contents.
            f = open(self.name, "rb")
            data = f.read()
            f.close()

        return data

    def create_initial_artifact(self):
        artifact = self.artifact_class.setup(self, self.name)
        if len(self.filters) == 0:
            artifact.is_last = True
            if artifact.final is None:
                artifact.final = True
        artifact.save()
        if artifact.binary_output:
            if artifact.doc.virtual:
                # make a fake file
                with open(artifact.filepath(), "wb") as f:
                    f.write("virtual file")
            else:
                shutil.copyfile(artifact.name, artifact.filepath())
        return artifact

    def run(self):
        self.step = 0
        start = time.time()
        time_start = start


        artifact = self.create_initial_artifact()
        self.artifacts.append(artifact)
        self.last_artifact = artifact
        artifact_key = artifact.key
        self.timing.append(("create-initial-artifact", time.time() - start))
        start = time.time()
        self.log.info("(step %s) [run] %s -> %s" % \
                 (self.step, artifact_key, artifact.filename()))

        for f in self.filters:
            previous_artifact = artifact
            artifact_key += "|%s" % f
            self.step += 1

            FilterClass = self.get_filter_for_alias(f)
            artifact = self.artifact_class.setup(self, artifact_key, FilterClass, previous_artifact)
            self.timing.append(("setup-step-%s" % self.step, time.time() - start))
            start=time.time()

            self.log.info("(starting step %s) [run] %s -> %s" % \
                     (self.step, artifact_key, artifact.filename()))

            artifact.run()
            self.timing.append(("run-step-%s" % self.step, time.time() - start))
            start = time.time()

            self.last_artifact = artifact
            self.artifacts.append(artifact)
            self.timing.append(("append-step-%s" % self.step, time.time() - start))
            start = time.time()

        self.last_artifact.is_last = True
        self.last_artifact.save_meta()
        self.db.append_artifacts(self.artifacts)
        self.timing.append(("append-artifacts", time.time() - start))
        start = time.time()

        # Make sure all additional inputs are saved.
        for k, a in artifact.inputs().iteritems():
            if not a.is_complete():
                if not a.additional:
                    # should only get here if this is an additional artifact
                    raise Exception("artifact %s should already be complete" % a.key)

                # set to complete and save
                a.state = 'complete'
                try:
                    a.save()
                    self.db.update_artifact(a)
                except IOError:
                    # no file was created, something went wrong in the script that should have
                    # generated this artifact
                    raise Exception("""You requested an additional file named %s
                    but its artifact file (%s) does not exist. Something may have gone wrong
                    in one of the filters of %s""" % (a.key, a.filepath(), a.created_by))

        self.timing.append(("save-additional-inputs", time.time() - start))
        self.elapsed = time.time() - time_start
        return self



from dexy.dexy_filter import DexyFilter
from StringIO import StringIO
import json
import os
import pycurl
import re
import urllib

class VanillaForumFilter(DexyFilter):
    ALIASES = ['vanilla']
    FORUM_CONFIG_FILE = 'vanilla-config.json'
    DISCUSSION_CONFIG_FILE = 'discuss.json|dexy'

    def load_forum_conf(self):
        if not os.path.exists(self.FORUM_CONFIG_FILE):
            raise Exception("Could not find config file called %s" %
                            self.FORUM_CONFIG_FILE)

        f = open(self.FORUM_CONFIG_FILE, "r")

        try:
            self.forum_conf = json.load(f)
        except ValueError as e:
            print "error parsing JSON in", self.FORUM_CONFIG_FILE
            raise e
        f.close()
        self.user_cookie = self.forum_conf['user-cookie']
        self.site = self.forum_conf['site'].rstrip('/')

    def load_discussion_conf(self):
        matches = [k for k in self.artifact.inputs().keys() if k.endswith(self.DISCUSSION_CONFIG_FILE)]
        try:
            self.k = matches[0]
        except IndexError as e:
            raise Exception("""no files matching %s were available as
                            inputs to %s""" % (self.DISCUSSION_CONFIG_FILE, self.artifact.key))

        try:
            self.discussion_conf = json.loads(self.artifact.inputs()[self.k]['data'])
        except ValueError as e:
            print "error parsing JSON in", self.k
            raise e

    def obtain_transient_key(self):
        c = pycurl.Curl()
        b = StringIO()
        c.setopt(c.URL, "%s/api/session" % self.site)
        c.setopt(c.COOKIE, "Vanilla=%s" % self.user_cookie)
        c.setopt(c.WRITEFUNCTION, b.write)
        c.perform()

        session_info = json.loads(b.getvalue())
        self.transient_key = session_info['user']['TransientKey']

    def process_text(self, input_text):
        self.load_forum_conf()
        self.load_discussion_conf()
        self.obtain_transient_key()

        if self.discussion_conf.has_key('Name'):
            discussion_name = self.discussion_conf['Name']
        else:
            raise Exception("Please specify a 'Name' for your discussion.")

        if self.discussion_conf.has_key('CategoryID'):
            discussion_category_id = self.discussion_conf['CategoryID']
        else:
            print "No 'CategoryID' specified, using default"
            discussion_category_id = 1


        discussion_data = [
          ("Discussion/Name", discussion_name),
          ("Discussion/CategoryID", discussion_category_id),
          ("Discussion/TransientKey", self.transient_key),
          ("Discussion/Body", input_text)
          ]

        if self.discussion_conf.has_key('DiscussionID'):
           discussion_data.append(('Discussion/DiscussionID', self.discussion_conf['DiscussionID']))
        else:
           print "no 'DiscussionID' specified, creating new discussion"

        c = pycurl.Curl()
        b = StringIO()
        c.setopt(c.URL, "%s/api/discussion/add" % self.site)
        c.setopt(c.COOKIE, "Vanilla=%s" % self.user_cookie)
        c.setopt(c.POST, 1)
        c.setopt(c.POSTFIELDS, urllib.urlencode(discussion_data))
        c.setopt(c.WRITEFUNCTION, b.write)
        c.perform()

        result = json.loads(b.getvalue())
        print result
        self.discussion_conf['DiscussionID'] = result['DiscussionID']

        json_file = re.sub('\|dexy$', "", self.k)
        f = open(json_file, 'w')
        json.dump(self.discussion_conf, f)
        f.close()
        return "ok"

class VanillaForumCommentFilter(VanillaForumFilter):
    ALIASES = ['vanillacomment']
    DISCUSSION_CONFIG_FILE = 'comment.json|dexy'

    def process_text(self, input_text):
        self.load_forum_conf()
        self.load_discussion_conf()

        self.obtain_transient_key()

        if self.discussion_conf.has_key('CategoryID'):
            comment_category_id = self.discussion_conf['CategoryID']
        else:
            print "No 'CategoryID' specified, using default"
            comment_category_id = 1

        if self.discussion_conf.has_key('DiscussionID'):
            comment_discussion_id = self.discussion_conf['DiscussionID']
        else:
            raise Exception("You must specify a valid DiscussionID")


        comment_data = [
          ("Comment/CategoryID", comment_category_id),
          ("Comment/DiscussionID", comment_discussion_id),
          ("Comment/TransientKey", self.transient_key),
          ("Comment/Body", input_text)
          ]

        if self.discussion_conf.has_key('CommentID'):
           comment_data.append(('Comment/CommentID', self.discussion_conf['CommentID']))
        else:
           print "no 'CommentID' specified, creating new comment"

        c = pycurl.Curl()
        b = StringIO()
        c.setopt(c.URL, "%s/api/comment/add" % self.site)
        c.setopt(c.COOKIE, "Vanilla=%s" % self.user_cookie)
        c.setopt(c.POST, 1)
        c.setopt(c.POSTFIELDS, urllib.urlencode(comment_data))
        c.setopt(c.WRITEFUNCTION, b.write)
        c.perform()

        result = json.loads(b.getvalue())
        print result
        self.discussion_conf['CommentID'] = result['CommentID']

        json_file = re.sub('\|dexy$', "", self.k)
        f = open(json_file, 'w')
        json.dump(self.discussion_conf, f)
        f.close()
        return "ok"



from dexy.dexy_filter import DexyFilter
from dexy.utils import print_string_diff
from dexy.utils import wrap_text
from ordereddict import OrderedDict
import json
import os
import re
import uuid

class TestFilter(DexyFilter):
    """
    Lets you test the output of running a filter, raises exception if expectation not met.

    If the test passes, the original content is returned so you can chain this filter.
    """

    ALIASES = ['test']

    def process(self):
        if self.artifact.controller_args['disabletests']:
            print "tests disabled, not running test", self.artifact.key
            self.artifact.data_dict = self.artifact.input_data_dict
            return False

        print "testing", self.artifact.key, "...",

        if self.artifact.args.has_key('test-expects'):
            # 'expects' should be an exact match
            expects = self.artifact.args['test-expects']
            if expects.startswith("@"):
                expects_file = open(expects.lstrip("@"), "r")
                expects_content = expects_file.read()
                expects_file.close()
            else:
                expects_content = expects
            expects = expects_content

            if hasattr(expects, 'items'):
                for k, v in self.artifact.input_data_dict.items():
                    msg = print_string_diff(expects[k], v)
                    msg += "\nexpected output '%s' (section %s) does not match actual '%s'" % (expects[k], k, v)
                    assert expects[k] == v, msg
            else:
                inp = self.artifact.input_text()
                msg = print_string_diff(expects, inp)
                msg += "\nexpected output '%s' does not match actual '%s'" % (expects, inp)
                assert expects == inp, msg
        elif self.artifact.args.has_key('test-includes'):
            # specified text should appear somewhere in the output
            includes_text = self.artifact.args['test-includes']
            assert includes_text in self.artifact.input_text()

        elif self.artifact.args.has_key('expects'):
            raise Exception("'expects' is deprecated, please use 'test-expects' instead.")

        else:
            raise Exception("Must pass one of 'test-expects', 'test-includes' to test filter so there is something to test.")

        comment = ""
        if self.artifact.args.has_key("comment"):
           comment = "[%s]" % self.artifact.args['comment']
        print "ok", comment

        # Return the original output, so 'test' filter doesn't change anything.
        self.artifact.data_dict = self.artifact.input_data_dict

class JoinFilter(DexyFilter):
    """
    Takes sectioned code and joins it into a single section. Some filters which
    don't preserve sections will raise an error if they receive multiple
    sections as input, so this forces acknowledgement that sections will be
    lost.
    """
    INPUT_EXTENSIONS = [".*"]
    OUTPUT_EXTENSIONS = [".*"]
    ALIASES = ['join']

    def process_dict(self, input_dict):
        return {'1' : self.artifact.input_text()}

class FooterFilter(DexyFilter):
    """
    Adds a footer to file. Looks for a file named _footer.ext where ext is the
    same extension as the file this is being applied to. So _footer.html for a
    file named index.html. Or, you can specify the name of the footer file to
    use in the args by passing a value to 'footer'.
    """
    INPUT_EXTENSIONS = [".*"]
    OUTPUT_EXTENSIONS = [".*"]
    ALIASES = ['ft', 'footer']
    DEFAULT_FOOTER_FILTERS = ['dexy', 'jinja']

    def process_text(self, input_text):
        footer_doc = None
        inputs = self.artifact.inputs()
        footer_input_keys = [k for k in self.artifact.inputs().keys() if "footer" in k]

        if self.artifact.args.has_key('footer'):
            # allow specifying footer keys
            footer_key = self.artifact.args['footer']
            if not footer_key in inputs:
                raise Exception("You requested footer %s, but this isn't an input to %s" % (footer_key, self.artifact.document_key))
            footer_doc = inputs[footer_key]

        else:
            # nothing specified, look for the default pattern
            footer_key = "_footer%s" % self.artifact.ext

            path_elements = self.artifact.name.split(os.sep)[:-1]
            footer_doc = None
            n = len(path_elements)

            for i in range(0, n+1):
                # Start in the immediate directory, proceed through parent
                # directories as far as project root until a footer file is
                # found.
                if i < n:
                    directory = os.path.join(*(path_elements[0:(n-i)]))
                    try_footer_in_dir = os.path.join(directory, footer_key)
                else:
                    try_footer_in_dir = footer_key

                if try_footer_in_dir in footer_input_keys:
                    footer_doc = inputs[try_footer_in_dir]
                else:
                    for pattern in self.DEFAULT_FOOTER_FILTERS:
                        if pattern:
                            try_key = "%s|%s" % (try_footer_in_dir, pattern)
                        if try_key in footer_input_keys:
                            footer_doc = inputs[try_key]
                            break

                if footer_doc:
                    break

        if not footer_doc:
            self.log.debug("keys for %s: %s" % (self.artifact.key, ", ".join(self.artifact._inputs.keys())))
            msg_str = "couldn't find a footer like %s as an input to %s (also tried with filters %s)"
            msg = msg_str % (footer_key, self.artifact.document_key, ", ".join(self.DEFAULT_FOOTER_FILTERS))
            raise Exception(msg)

        self.log.debug("using %s as footer for %s" % (footer_doc.key, self.artifact.document_key))
        footer_text = footer_doc.output_text()
        return "%s\n%s" % (input_text, footer_text)

class HeaderFilter(DexyFilter):
    """
    Adds a header to file. Looks for a file named _header.ext where ext is the
    same extension as the file this is being applied to. So _header.html for a
    file named index.html. Or, you can specify the name of the header file to
    use in the args by passing a value to 'header'.
    """
    INPUT_EXTENSIONS = [".*"]
    OUTPUT_EXTENSIONS = [".*"]
    ALIASES = ['hd', 'header']
    DEFAULT_HEADER_FILTERS = ['dexy', 'jinja']

    def process_text(self, input_text):
        header_doc = None
        inputs = self.artifact.inputs()

        if self.artifact.args.has_key('header'):
            # allow specifying header keys
            header_key = self.artifact.args['header']
            if not header_key in inputs:
                raise Exception("You requested header %s, but this isn't an input to %s" % (header_key, self.artifact.document_key))
            header_doc = inputs[header_key]

        else:
            # nothing specified, look for the default pattern
            header_key = "_header%s" % self.artifact.ext

            path_elements = self.artifact.name.split(os.sep)[:-1]
            header_doc = None
            n = len(path_elements)

            for i in range(0, n+1):
                # Start in the immediate directory, proceed through parent
                # directories as far as project root until a header file is
                # found.
                if i < n:
                    directory = os.path.join(*(path_elements[0:(n-i)]))
                    try_header_in_dir = os.path.join(directory, header_key)
                else:
                    try_header_in_dir = header_key

                if inputs.has_key(try_header_in_dir):
                    header_doc = inputs[try_header_in_dir]
                else:
                    for pattern in self.DEFAULT_HEADER_FILTERS:
                        if pattern:
                            try_key = "%s|%s" % (try_header_in_dir, pattern)
                        if inputs.has_key(try_key):
                            header_doc = inputs[try_key]
                            break

                if header_doc:
                    break

        if not header_doc:
            self.log.debug("keys for %s: %s" % (self.artifact.key, ", ".join(self.artifact._inputs.keys())))
            msg_str = "couldn't find a header like %s as an input to %s (also tried with filters %s)"
            msg = msg_str % (header_key, self.artifact.document_key, ", ".join(self.DEFAULT_HEADER_FILTERS))
            raise Exception(msg)

        self.log.debug("using %s as header for %s" % (header_doc.key, self.artifact.document_key))
        return "%s\n%s" % (header_doc.output_text(), input_text)

class HeadFilter(DexyFilter):
    """
    Returns just the first 10 lines of input.
    """
    ALIASES = ['head']
    def process_text(self, input_text):
        return "\n".join(input_text.split("\n")[0:10]) + "\n"

class WordWrapFilter(DexyFilter):
    """
    Wraps text after 79 characters (tries to preserve existing line breaks and
    spaces).
    """
    ALIASES = ['ww', 'wrap']

    def process_text(self, input_text):
        return wrap_text(input_text, 79)

class SplitHtmlFilter(DexyFilter):
    """
    Create multiple HTML pages from a single template, with an automatic index page.

    The split filter looks for specially formatted HTML comments in your
    document and splits your HTML into separate pages at each split comment.
    """
    ALIASES = ['split', 'splithtml']
    INPUT_EXTENSIONS = [".html"]
    OUTPUT_EXTENSIONS = [".html"]
    FINAL = True

    def process(self):
        parent_dir = os.path.dirname(self.artifact.canonical_filename())
        input_text = self.artifact.input_text()

        if input_text.find("<!-- endsplit -->") > 0:
            body, footer = re.split("<!-- endsplit -->", input_text, maxsplit=1)
            sections = re.split("<!-- split \"(.+)\" -->\n", body)
            header = sections[0]

            pages = OrderedDict()
            index_content = None
            for i in range(1, len(sections), 2):
                if sections[i] == 'index':
                    index_content = sections[i+1]
                else:
                    section_name = sections[i]
                    # TODO proper url/filename escaping
                    section_url = section_name.replace(" ","-")

                    filename = "%s.html" % section_url
                    filepath = os.path.join(parent_dir, filename)
                    pages[section_name] = filename

                    artifact = self.artifact.add_additional_artifact(filepath, 'html')
                    artifact.set_data(header + sections[i+1] + footer)
                    artifact.save()

                    self.artifact.log.debug("added key %s to artifact %s ; links to file %s" %
                              (filepath, self.artifact.key, artifact.filename()))

            index_items = []
            for k in sorted(pages.keys()):
                index_items.append("""<li><a href="%s">%s</a></li>""" %
                                   (pages[k], k))

            output_dict = OrderedDict()
            output_dict['header'] = header
            if index_content:
                output_dict['index-page-content'] = index_content

            if self.artifact.args.has_key("split-ul-class"):
                ul = "<ul class=\"%s\">" % self.artifact.args['split-ul-class']
            else:
                ul = "<ul class=\"split\">"

            output_dict['index'] = "%s\n%s\n</ul>" % (ul, "\n".join(index_items))
            output_dict['footer'] = footer
        else:
            # No endsplit found, do nothing.
            output_dict = self.artifact.input_data_dict
        self.artifact.data_dict = output_dict

class SplitLatexFilter(DexyFilter):
    """Splits a latex doc into multiple latex docs."""
    ALIASES = ['splitlatex']
    INPUT_EXTENSIONS = [".tex"]
    OUTPUT_EXTENSIONS = [".tex"]
    FINAL = True

    def process(self):
        parent_dir = os.path.dirname(self.artifact.canonical_filename())
        input_text = self.artifact.input_text()

        if input_text.find("%% endsplit\n") > 0:
            body, footer = re.split("%% endsplit\n", input_text, maxsplit=1)
            sections = re.split("%% split \"(.+)\"\n", body)
            header = sections[0]

            pages = OrderedDict()
            for i in range(1, len(sections), 2):
                section_name = sections[i]
                # TODO proper url/filename escaping
                section_url = section_name.replace(" ","-")
                source = header + sections[i+1] + footer

                ext = '.tex'

                filename = "%s%s" % (section_url, ext)
                filepath = os.path.join(parent_dir, filename)
                pages[section_name] = filename

                artifact = self.artifact.__class__(filepath)
                artifact.ext = ext
                artifact.binary = False
                artifact.final = True
                artifact.artifacts_dir = self.artifact.artifacts_dir
                artifact.hashstring = str(uuid.uuid4())
                artifact.set_data(source)
                artifact.save()

                self.artifact.inputs()[filepath] = artifact
                self.artifact.log.debug("added key %s to artifact %s ; links to file %s" %
                          (filepath, self.artifact.key, artifact.filename()))

            index_items = []
            for k in sorted(pages.keys()):
                index_items.append("""<li><a href="%s">%s</a></li>""" %
                                   (pages[k], k))

        output_dict = self.artifact.input_data_dict
        self.artifact.data_dict = output_dict

class PrettyPrintJsonFilter(DexyFilter):
    ALIASES = ['ppjson']
    OUTPUT_EXTENSIONS = ['.json']

    def process_text(self, input_text):
        json_content = json.loads(input_text)
        return json.dumps(json_content, sort_keys=True, indent=4)

class SillyFilter(DexyFilter):
    ALIASES =['silly']

    def process_text(self, input_text):
        return "you said: '%s'\n that's silly!\n" % input_text

class SectionsByLineFilter(DexyFilter):
    ALIASES = ['lines']

    def process_text_to_dict(self, input_text):
        data_dict = OrderedDict()
        for i, line in enumerate(input_text.splitlines()):
            data_dict["%s" % (i+1)] = line
        return data_dict



from dexy.artifacts.file_system_pickle_artifact import FileSystemPickleArtifact
import cPickle

class FileSystemcPickleArtifact(FileSystemPickleArtifact):
    """Artifact which persists data by writing to the file system and using
    cjson for serializing metadata"""
    def write_dict_to_file(self, data_dict, filepath):
        with open(filepath, "wb") as f:
            cPickle.dump(data_dict, f)

    def read_dict_from_file(self, filepath):
        with open(filepath, "rb") as f:
            return cPickle.load(f)



from dexy.dexy_filter import DexyFilter
import shutil

class DeprecatedCopyFilter(DexyFilter):
    INPUT_EXTENSIONS = [".*"]
    OUTPUT_EXTENSIONS = [".*"]
    ALIASES = ['cp']
    BINARY = True
    FINAL = True

    def process(self):
        print self.artifact.key, "- The 'cp' filter is deprecated. This filter is no longer necessary, you can remove '|cp' from your specification and the file will still be copied."
        shutil.copyfile(self.artifact.previous_artifact_filepath, self.artifact.filepath())



import Cheetah.Template
import dexy.filters.templating_filters

class CheetahFilter(dexy.filters.templating_filters.TemplateFilter):
    ALIASES = ['cheetah']

    def process_text(self, input_text):
        return str(Cheetah.Template.Template(input_text, searchList = self.run_plugins()))



from dexy.dexy_filter import DexyFilter
import textile

class TextileFilter(DexyFilter):
    INPUT_EXTENSIONS = ['.txt', '.textile']
    OUTPUT_EXTENSIONS = ['.html']
    ALIASES = ['pytextile']

    def process_text(self, input_text):
        return textile.textile(input_text)




from dexy.dexy_filter import DexyFilter
from ordereddict import OrderedDict
from pynliner import Pynliner

class PynlinerFilter(DexyFilter):
    """
    Move CSS inline, for posting to web without a stylesheet or for emailing.
    For now hard-coded to look for a pastie.css file in project root.
    """
    ALIASES = ['pynliner', 'inlinecss']

    def process_dict(self, input_dict):
        #matches = [k for k in self.artifact.input_artifacts_dict.keys() if k.endswith(".css|dexy")]
        #k = matches[0]
        css = open("pastie.css", "r").read()

        output_dict = OrderedDict()
        for k, v in input_dict.items():
            try:
                p = Pynliner(self.log)
            except TypeError:
                print "the pynliner filter says: please install pynliner from source (https://github.com/rennat/pynliner.git) or version > 0.2.1"
                p = Pynliner()
            p.from_string(v).with_cssString(css)
            output_dict[k] = p.run()
        return output_dict



from dexy.reporter import Reporter
import datetime
import os
import shutil
import tarfile

class TarzipOutputReporter(Reporter):
    """
    Reporter which creates a .tgz of all output files. Follows same naming
    conventions as OutputReporter.
    """
    SUBDIR_NAME = "dexy-output"

    def run(self):
        timestamp = datetime.datetime.now().strftime("%Y-%m-%d--%H-%M-%S")
        report_filename = os.path.join(self.logsdir, "output-%s.tgz" % timestamp)
        tar = tarfile.open(report_filename, mode="w:gz")

        self.load_batch_artifacts()
        for artifact in self.artifacts.values():
            if (artifact.is_last and artifact.final) or artifact.additional:
                arcname = os.path.join(self.SUBDIR_NAME, artifact.canonical_filename())
                tar.add(artifact.filepath(), arcname=arcname)

        tar.close()
        shutil.copyfile(report_filename, os.path.join(self.logsdir, "output-latest.tgz"))

class InSituReporter(Reporter):
    """
    InSitu Reporter creates files directly in your project. Useful for using
    dexy to create files like README files which need to live in a certain
    place in your project. Templates should have a different file extension
    than the final file. InSitu Reporter will not overwrite existing files,
    delete any previously generated files before running this.
    """
    ALLREPORTS = False
    def run(self):
        self.load_batch_artifacts()
        for artifact in self.artifacts.values():
            if (artifact.is_last and artifact.final) or artifact.additional:
                fn = artifact.canonical_filename()
                if os.path.exists(fn):
                    print "InSituReporter not overwriting existing file", fn, "please delete this file before running InSitu reporter"
                else:
                    artifact.write_to_file(fn)

class OutputReporter(Reporter):
    """
    Saves dexy-processed files in a directory under their canonical filenames.
    Tries to be smart about which files you actually care about and want to
    see. See LongOutputReporter if you want to see everything (with uglier names).
    """
    REPORTS_DIR = 'output'

    def run(self):
        self.create_reports_dir(self.REPORTS_DIR)
        self.load_batch_artifacts()
        for artifact in self.artifacts.values():
            if (artifact.is_last and artifact.final) or artifact.additional:
                fn = artifact.canonical_filename()
                fp = os.path.join(self.REPORTS_DIR, fn)
                if os.path.exists(fp):
                    self.log.warn("two or more final artifacts have canonical path %s" % fp)
                    self.log.warn("most recent is %s" % artifact.key)
                else:
                    self.log.debug("saving %s to %s" % (artifact.key, fn))
                artifact.write_to_file(fp)

class LongOutputReporter(Reporter):
    """
    Saves all dexy-processed files in a directory using unique filenames.
    """
    REPORTS_DIR = 'output-long'
    def run(self):
        self.create_reports_dir(self.REPORTS_DIR)
        self.load_batch_artifacts()
        for artifact in self.artifacts.values():
            if artifact.is_last:
                fn = artifact.long_canonical_filename()
                fp = os.path.join(self.REPORTS_DIR, fn)
                artifact.write_to_file(fp)



from dexy.artifacts.file_system_json_artifact import FileSystemJsonArtifact
import cjson

class FileSystemCjsonArtifact(FileSystemJsonArtifact):
    """Artifact which persists data by writing to the file system and using
    cjson for serializing metadata"""

    def write_dict_to_file(self, data_dict, filepath):
        with open(filepath, "wb") as f:
            f.write(cjson.encode(data_dict))

    def read_dict_from_file(self, filepath):
        with open(filepath, "rb") as f:
            return cjson.decode(f.read())



from dexy.constants import Constants
import dexy.artifact
import dexy.database
import dexy.introspect
import dexy.utils
import os
import shutil

class Reporter(object):
    ALLREPORTS = True # whether to be included in 'allreports', set to false for reporters with side effects
    REPORTS_DIR = None
    LDIR = Constants.DEFAULT_LDIR
    LFILE = Constants.DEFAULT_LFILE
    ACLASS = Constants.DEFAULT_ACLASS
    DBFILE = Constants.DEFAULT_DBFILE
    DBCLASS = Constants.DEFAULT_DBCLASS
    SAFETY_FILENAME = ".dexy-generated"
    README_FILENAME = "README"

    def create_reports_dir(self, reports_dir):
        safety_filepath = os.path.join(reports_dir, self.SAFETY_FILENAME)
        readme_filepath = os.path.join(reports_dir, self.README_FILENAME)

        if os.path.exists(reports_dir) and not os.path.exists(safety_filepath):
            raise Exception("Please remove directory %s, Dexy wants to put a report here but doesn't want to overwrite anything by accident." % os.path.abspath(reports_dir))
        elif os.path.exists(reports_dir):
            shutil.rmtree(reports_dir)

        os.mkdir(reports_dir)

        with open(safety_filepath, "w") as f:
            f.write("""
            This directory was generated by the %s Dexy Reporter and
            may be deleted without notice.\n\n""" % self.__class__.__name__)
        with open(readme_filepath, "w") as f:
            f.write("""
            This directory was generated by the %s Dexy Reporter and
            may be deleted without notice.\n\n""" % self.__class__.__name__)

    def __init__(self, logsdir=LDIR, logfile=LFILE, artifact_class=ACLASS, batch_id=None,
            dbfile=DBFILE, controller=None, dbclass=DBCLASS, hashfunction='md5'):

        self.db = dexy.utils.get_db(dbclass, logsdir=logsdir, dbfile=dbfile)

        if batch_id:
            self.batch_id = batch_id
        else:
            self.batch_id = self.db.max_batch_id()

        self.dbfile = dbfile
        self.log = dexy.utils.get_log('dexy.controller', logsdir=logsdir, logfile=logfile)
        self.log.debug("Reporter %s initialized" % self.__class__.__name__)
        self.logfile = logfile
        self.logsdir = logsdir
        self.hashfunction = hashfunction

        if isinstance(artifact_class, str):
            artifact_classes = dexy.introspect.artifact_classes()
            self.artifact_class = artifact_classes[artifact_class]
        elif issubclass(artifact_class, dexy.artifact.Artifact):
            self.artifact_class = artifact_class
        else:
            raise Exception("expected artifact_class to be class name or class, got %s"  % type(artifact_class))

        if controller:
            self.batch_info = controller.batch_info()
        elif self.batch_id:
            self.batch_info = dexy.utils.load_batch_info(self.batch_id, logsdir)
        else:
            print "warning, no batch id could be detected"
            self.batch_info = None

    def load_batch_artifacts(self):
        db = dexy.utils.get_db(logsdir=self.logsdir, dbfile=self.dbfile)
        refs = db.references_for_batch_id(self.batch_id)
        self.batch_refs = refs
        self.artifacts = dict((r['hashstring'], self.artifact_class.retrieve(r['hashstring'], self.hashfunction)) for r in refs)

    def run(self):
        pass



from dexy.dexy_filter import DexyFilter
from idiopidae.runtime import Composer
from pygments import highlight
from pygments.formatters.html import HtmlFormatter
from pygments.formatters.latex import LatexFormatter
from pygments.lexers.agile import PythonLexer
from dexy.introspect import INSTALL_DIR
import idiopidae.parser
import inspect
import json
import os
import pkgutil
import shutil
import subprocess
import sys

"""
Filters that parse and process various language's documentation systems to make
this information available in Dexy documents. Filters work by processing a
config file that specifies which libraries should be processed. These filters
work for languages where documentation can be generated by referencing
installed libraries, rather than needing source code.
"""

class PythonDocumentationFilter(DexyFilter):
    ALIASES = ["pydoc"]
    INPUT_EXTENSIONS = [".txt"]
    OUTPUT_EXTENSIONS = [".json"]
    COMPOSER = Composer()
    LEXER = PythonLexer()
    LATEX_FORMATTER = LatexFormatter()
    HTML_FORMATTER = HtmlFormatter()

    def fetch_item_content(self, cm):
        is_method = inspect.ismethod(cm)
        is_function = inspect.isfunction(cm)

        if is_method or is_function:
            try:
                source = inspect.getsource(cm)
            except IOError:
                source = ""

            builder = idiopidae.parser.parse('Document', source + "\n\0")
            sections = {}

            for i, s in enumerate(builder.sections):
                lines = builder.statements[i]['lines']
                sections[s] = self.COMPOSER.format(lines, self.LEXER, self.HTML_FORMATTER)

            if len(sections) == 1:
                return sections.values()[0]
            else:
                return sections
        else:
            try:
                # If this can be JSON-serialized, leave it alone...
                json.dumps(cm)
                return cm
            except TypeError:
                # ... if it can't, convert it to a string to avoid problems.
                return str(cm)

    def process_text(self, input_text):
        """
        input_text should be a list of installed python libraries to document.
        """
        package_names = input_text.split()
        packages = [__import__(package_name) for package_name in package_names]
        packages_info = {}

        for package in packages:
            package_name = package.__name__
            method_source_code = {}
            class_info = {}
            prefix = package.__name__ + "."
            for module_loader, name, ispkg in pkgutil.walk_packages(package.__path__, prefix=prefix):
                try:
                    __import__(name)
                    mod = sys.modules[name]

                    for k, m in inspect.getmembers(mod):
                        if not inspect.isclass(m) and hasattr(m, '__module__') and m.__module__.startswith(package_name):
                            # TODO figure out how to get module constants
                            key = "%s.%s" % (m.__module__, k)
                            item_content = self.fetch_item_content(m)
                            method_source_code[key] = item_content

                        elif inspect.isclass(m) and m.__module__.startswith(package_name):
                            class_key = "%s.%s" % (name, k)
                            class_info[class_key] = {}
                            try:
                                class_info[class_key]['source'] = highlight(inspect.getsource(m), self.LEXER, self.HTML_FORMATTER)
                            except IOError:
                                print "can't get source for", class_key
                                class_info[class_key]['source'] = ""

                            for ck, cm in inspect.getmembers(m):
                                key = "%s.%s.%s" % (name, k, ck)
                                item_content = self.fetch_item_content(cm)
                                method_source_code[key] = item_content
                                class_info[class_key][ck] = item_content
                except ImportError as e:
                    print e
            packages_info[package.__name__] = method_source_code
        return json.dumps(packages_info, indent=4)

class RDocumentationFilter(DexyFilter):
    """
    Can be run on a text file listing packages to be processed, or an R script
    which should define a list of package names (strings) named 'packages', the
    latter option so that you can include some R code prior to automated code
    running.
    """
    ALIASES = ["rdoc"]
    INPUT_EXTENSIONS = [".txt", ".R"]
    OUTPUT_EXTENSIONS = [".json"]

    def process(self):
        # Create a temporary directory to run R in.
        self.artifact.create_temp_dir()
        td = self.artifact.temp_dir()

        r_script_file = os.path.join(INSTALL_DIR, 'dexy', 'ext', "introspect.R")
        self.log.debug("script file: %s" % r_script_file)

        with open(r_script_file, "r") as f:
            r_script_contents = f.read()

        if self.artifact.input_ext == ".txt":
            # A text file containing the names of packages to process.
            package_names = self.artifact.input_text().split()
            script_start = "packages <- c(%s)" % ",".join("\"%s\"" % n for n in package_names)
        elif self.artifact.input_ext == ".R":
            script_start = self.artifact.input_text()
        else:
            raise Exception("Unexpected input file extension %s" % self.artifact.input_ext)

        script_filename = os.path.join(td, "script.R")

        with open(script_filename, "w") as f:
            f.write(script_start + "\n")
            f.write(r_script_contents)

        command = "R --slave --vanilla < script.R"
        self.log.debug("About to run %s" % command)

        proc = subprocess.Popen(command, shell=True,
                                cwd=td,
                                stdin=subprocess.PIPE,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                )
        stdout, stderr = proc.communicate()
        self.artifact.stdout = stdout
        shutil.copyfile(os.path.join(td, "dexy--r-doc-info.json"), self.artifact.filepath())




from dexy.reporter import Reporter
from jinja2 import Environment
from jinja2 import FileSystemLoader
import datetime
import os
import shutil
import this_report_is_deprecated_until_it_gets_updated

class SiteReporter(Reporter):
    def setup_template(self):
        env = Environment()
        env.loader = FileSystemLoader(os.path.dirname(__file__))
        self.template = env.get_template('site_reporter_template.html')

    def setup_final_artifacts(self):
        self.artifacts = []

        for doc in self.controller.docs:
            artifact = doc.final_artifact()

            if artifact.final:
                self.artifacts.append(artifact)

            for k, a in artifact._inputs.items():
                if not a:
                    print "no artifact exists for key", k
                else:
                    if a.additional and a.final:
                        if not a.is_complete():
                            a.state = 'complete'
                            a.save()
                        self.artifacts.append(artifact)

    def setup_dirs(self):
        """Return a dict of directories and the artifacts in each directory."""
        self.dirs = {}
        for artifact in self.artifacts:
            dirname = os.path.dirname(artifact.canonical_filename())
            if self.dirs.has_key(dirname):
                if not artifact.key in [a.key for a in self.dirs[dirname]]:
                    self.dirs[dirname].append(artifact)
            else:
                self.dirs[dirname] = [artifact]

        self.html_filenames = {}
        self.artifact_filenames = {}
        for d in self.dirs.keys():
            self.html_filenames[d] = {}
            self.artifact_filenames[d] = {}

    def write_html_page_and_artifact(self, artifact):
        """Write artifact and maybe a HTML display page."""

        artifact_fn = artifact.canonical_filename()
        html_page_fn = "%s.html" % artifact_fn
        if artifact.ext == '.html':
            if not "|hd" in artifact.key: # TODO more robust test of whether full or fragment HTML
                artifact_fn = False
            elif os.path.basename(artifact_fn) == 'index.html':
                # we need index.html to be the html page
                html_page_fn = artifact_fn
                artifact_fn = artifact.long_canonical_filename()

        if artifact_fn:
            artifact.write_to_file(os.path.join(self.report_dir, artifact_fn))

        if artifact.ext in ['.html']:
            content = artifact.output_text()
        elif artifact.ext in ['.txt']:
            content = """<pre>\n%s\n</pre>""" % artifact.output_text()
        elif artifact.ext in ['.png', '.jpg', '.gif']:
            content = """<a href="%s"><img src="%s" /></a>""" % (artifact_fn, artifact_fn)
        elif artifact.ext in ['', '.pdf', '.css', '.rb', '.py', '.swf', '.R']:
            content = """<a href="%s">%s</a>""" % (artifact_fn, artifact_fn)
        else:
            self.log.debug("site reporter using default handling for extension %s" % artifact.ext)
            content = """<a href="%s">%s</a>""" % (artifact_fn, artifact_fn)

        env_data = {
            'a' : artifact,
            'title' : artifact.key,
            'content' : content,
            'directory' : os.path.dirname(html_page_fn),
            'dirs' : self.dirs.keys(),
            'pretty_site_name' : self.pretty_site_name,
            'relpath' : os.path.relpath,
            'basename' : os.path.basename,
            'site_name' : self.site_name,
            'artifact_fn' : artifact_fn,
            'html_page_fn' : html_page_fn
        }
        html_page_parent_dir = os.path.dirname(os.path.join(self.report_dir, html_page_fn))
        if not os.path.exists(html_page_parent_dir):
                os.makedirs(html_page_parent_dir)
        self.template.stream(env_data).dump(os.path.join(self.report_dir, html_page_fn))

    def run(self):
        self.timestamp = datetime.datetime.now().strftime("%Y-%m-%d--%H-%M-%S")
        self.report_dir = os.path.join(self.controller.logs_dir, "site-%s" % self.timestamp)
        self.latest_report_dir = os.path.join(self.controller.logs_dir, "site-latest")
        self.site_name = os.path.split(os.path.abspath(os.curdir))[1]
        self.pretty_site_name = " ".join([n.capitalize() for n in self.site_name.split("-")])

        # This shouldn't matter unless dexy gets run twice in 1 second
        shutil.rmtree(self.report_dir, ignore_errors=True)
        os.mkdir(self.report_dir)
        self.setup_template()
        self.setup_final_artifacts()
        self.setup_dirs()

        for artifact in self.artifacts:
            self.write_html_page_and_artifact(artifact)

        def write_index(directory, children):
            # create an index page for each directory
            index_filename = os.path.join(self.report_dir, directory, 'index.html')
            env_data = {
                'children' : children,
                'directory' : directory,
                'content' : '',
                'dirs' : self.dirs.keys(),
                'pretty_site_name' : self.pretty_site_name,
                'relpath' : os.path.relpath,
                'basename' : os.path.basename,
                'site_name' : self.site_name
            }
            self.template.stream(env_data).dump(index_filename)


        for dirname, children in self.dirs.iteritems():
            write_index(dirname, children)

        shutil.rmtree(self.latest_report_dir, ignore_errors=True)
        shutil.copytree(self.report_dir, self.latest_report_dir)



from dexy.filters.process_filters import ProcessFilter
from ordereddict import OrderedDict
import os
import pexpect
import re

class PexpectReplFilter(ProcessFilter):
    """
    Base class for filters which use pexpect to retrieve output line-by-line based on detecting prompts.
    """
    ALIASES = ['pexpectreplfilter']
    ALLOW_MATCH_PROMPT_WITHOUT_NEWLINE = False
    INITIAL_PROMPT = None
    LINE_ENDING = "\r\n"
    PROMPTS = ['>>>', '...'] # Python uses >>> prompt normally and ... when in multi-line structures like loops
    PROMPT_REGEX = None
    PS1 = None
    PS2 = None
    PS3 = None
    PS4 = None
    SAVE_VARS_TO_JSON_CMD = None
    TRIM_PROMPT = '>>>'

    def prompt_search_terms(self):
        """
        Search first for the prompt (or prompts) following a line ending.
        Also optionally allow matching the prompt with no preceding line ending.
        """

        if self.PROMPT_REGEX:
            prompts = [self.PROMPT_REGEX]
        elif hasattr(self, 'PROMPT'):
            prompts = [self.PROMPT]
        else:
            prompts = self.PROMPTS

        if self.ALLOW_MATCH_PROMPT_WITHOUT_NEWLINE:
            return ["%s%s" % (self.LINE_ENDING, p) for p in prompts] + prompts
        else:
            return ["%s%s" % (self.LINE_ENDING, p) for p in prompts]

    def lines_for_section(self, section_text):
        """
        Take the section text and split it into lines which will be sent to the
        interpreter. This can be overridden if lines need to be defined
        differently, or if you don't want the extra newline at the end.
        """
        return section_text.splitlines() + ["\n"]

    def strip_trailing_prompts(self, section_transcript):
        lines = section_transcript.splitlines()
        while len(lines) > 0 and re.match("^\s*(%s)\s*$|^\s*$" % self.TRIM_PROMPT, lines[-1]):
            lines = lines[0:-1]
        return self.LINE_ENDING.join(lines)

    def strip_newlines(self, line):
        return line.replace(" \r", "")

    def process_dict(self, input_dict):
        output_dict = OrderedDict()

        # If we want to automatically record values of local variables in the
        # script we are running, we add a section at the end of script
        do_record_vars = self.artifact.args.has_key('record-vars') and self.artifact.args['record-vars']
        if do_record_vars:
            if not self.SAVE_VARS_TO_JSON_CMD:
                raise Exception("Can't record vars since SAVE_VARS_TO_JSON_CMD not set.")
            artifact = self.artifact.add_additional_artifact(self.artifact.key + "-vars.json", 'json')
            self.log.debug("Added additional artifact %s (hashstring %s) to store variables" % (artifact.key, artifact.hashstring))
            section_text = self.SAVE_VARS_TO_JSON_CMD % artifact.filename()
            input_dict['dexy--save-vars'] = section_text
            # TODO this may cause unexpected results if original script only had 1 section

        search_terms = self.prompt_search_terms()

        env = self.setup_env()
    	if not env:
    	    env = os.environ

        if env.has_key('PS1') and self.PS1:
            env['PS1'] = self.PS1
        if env.has_key('PS2') and self.PS2:
            env['PS2'] = self.PS2
        if env.has_key('PS3') and self.PS3:
            env['PS3'] = self.PS3
        if env.has_key('PS4') and self.PS4:
            env['PS4'] = self.PS4

        timeout = self.setup_timeout()

        # Spawn the process
        proc = pexpect.spawn(
                self.executable(),
                cwd=self.artifact.artifacts_dir,
                env=env)

        # Capture the initial prompt
        if self.INITIAL_PROMPT:
            proc.expect(self.INITIAL_PROMPT, timeout=timeout)
        elif self.PROMPT_REGEX:
            proc.expect(search_terms, timeout=timeout)
        else:
            proc.expect_exact(search_terms, timeout=timeout)

        start = proc.before + proc.after
        for section_key, section_text in input_dict.items():
            section_transcript = start
            start = ""

            lines = self.lines_for_section(section_text)
            for l in lines:
                section_transcript += start
                proc.send(l.rstrip() + "\n")
                if self.PROMPT_REGEX:
                    proc.expect(search_terms, timeout=timeout)
                else:
                    proc.expect_exact(search_terms, timeout=timeout)
                section_transcript += self.strip_newlines(proc.before)
                start = proc.after

            # Save this section's output
            output_dict[section_key] = self.strip_trailing_prompts(section_transcript)

        try:
            proc.close()
        except pexpect.ExceptionPexpect:
            raise Exception("process %s may not have closed" % proc.pid)

        if proc.exitstatus:
            self.handle_subprocess_proc_return(proc.exitstatus, str(output_dict))

        return output_dict

class RubyPexpectReplFilter(PexpectReplFilter):
    ALIASES = ['irb', 'rbrepl']
    EXECUTABLE = 'irb --simple-prompt'
    IGNORE_ERRORS = True
    INITIAL_PROMPT = "^>>"
    INPUT_EXTENSIONS = [".txt", ".rb"]
    OUTPUT_EXTENSIONS = [".rbcon"]
    PROMPTS = [">>", "?>"]
    TRIM_PROMPT = '>>'
    VERSION_COMMAND = 'irb --version'

class PythonPexpectReplFilter(PexpectReplFilter):
    ALIASES = ['pycon', 'pyrepl']
    EXECUTABLE = 'python'
    INPUT_EXTENSIONS = [".txt", ".py"]
    OUTPUT_EXTENSIONS = [".pycon"]
    TAGS = ['python', 'interpreter', 'language']
    VERSION_COMMAND = 'python --version'

    SAVE_VARS_TO_JSON_CMD = """
import json
dexy__vars_file = open("%s", "w")
dexy__x = {}
for dexy__k, dexy__v in locals().items():
    dexy__x[dexy__k] = str(dexy__v)

json.dump(dexy__x, dexy__vars_file)
dexy__vars_file.close()
"""

class IpythonPexpectReplFilter(PythonPexpectReplFilter):
    # TODO test for version of ipython that supports --classic
    ALIASES = ['ipython']
    EXECUTABLE = 'ipython --classic'
    IGNORE_ERRORS = True # TODO try to figure out why we are getting nonzero exit codes
    INPUT_EXTENSIONS = [".txt", ".py"]
    OUTPUT_EXTENSIONS = [".pycon"]
    VERSION_COMMAND = 'ipython -Version'

# untested below here

class RPexpectReplFilter(PexpectReplFilter):
    ALIASES = ['r', 'rint']
    EXECUTABLE = "R --quiet --vanilla"
    INPUT_EXTENSIONS = ['.txt', '.r', '.R']
    OUTPUT_EXTENSIONS = ['.Rout']
    PROMPTS = [">", "+"]
    TAGS = ['r', 'interpreter', 'language']
    TRIM_PROMPT = ">"
    INITIAL_PROMPT = "^>"
    VERSION_COMMAND = "R --version"
    SAVE_VARS_TO_JSON_CMD = """
if ("rjson" %%in%% installed.packages()) {
    library(rjson)
    dexy__json_file <- file("%s", "w")
    writeLines(toJSON(as.list(environment())), dexy__json_file)
    close(dexy__json_file)
} else {
   cat("Can't automatically save environment to JSON since rjson package not installed.")
}
"""

class RhinoInteractiveFilter(PexpectReplFilter):
    """
    Runs rhino JavaScript interpeter.
    """
    EXECUTABLE = "rhino"
    INPUT_EXTENSIONS = [".js", ".txt"]
    OUTPUT_EXTENSIONS = [".jscon"]
    ALIASES = ['jsint', 'rhinoint']
    PROMPTS = ['js>', '  >']
    TRIM_PROMPT = "js>"

class KshInteractiveFilter(PexpectReplFilter):
    """
    Runs ksh. Use to run bash scripts.
    """
    ALIASES = ['shint']
    EXECUTABLE = "ksh -i -r -e"
    INPUT_EXTENSIONS = [".txt", ".sh"]
    OUTPUT_EXTENSIONS = ['.sh-session']
    INITIAL_PROMPT = "^(#|\$)"
    PROMPTS = ["$", "#"]
    TRIM_PROMPT = "\$|#"
    PS1 = "$"

class ZshInteractiveFilter(PexpectReplFilter):
    """
    Runs zsh.
    """
    ALIASES = ['zshint']
    EXECUTABLE = "zsh --interactive --restricted --errexit"
    INPUT_EXTENSIONS = [".txt", ".sh"]
    OUTPUT_EXTENSIONS = ['.sh-session']
    PROMPT_REGEX = "[a-z\.]+(%|>)"
    ALLOW_MATCH_PROMPT_WITHOUT_NEWLINE = True
    IGNORE_ERRORS = True

class ClojureInteractiveFilter(PexpectReplFilter):
    """
    Runs clojure.
    """
    EXECUTABLE = 'clojure -r'
    INPUT_EXTENSIONS = [".clj", ".txt"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['clj', 'cljint']
    PROMPT = "user=> "

    def lines_for_section(self, input_text):
        input_lines = []
        current_line = []
        in_indented_block = False
        for l in input_text.splitlines():
            if re.match("^\s+", l):
                in_indented_block = True
                current_line.append(l)
            else:
                if len(current_line) > 0:
                    input_lines.append("\n".join(current_line))
                if in_indented_block:
                    # we have reached the end of this indented block
                    in_indented_block = False
                current_line = [l]
        input_lines.append("\n".join(current_line))
        return input_lines



from dexy.artifact import Artifact
from dexy.databases.sqlite_database import SqliteDatabase

def test_create_table():
    db = SqliteDatabase(dbfile=None)
    assert len(db.conn.execute("select * from sqlite_master where type='table' and name='artifacts'").fetchall()) == 1
    print db.create_table_sql()

def test_max_batch_id():
    db = SqliteDatabase(dbfile=None)
    assert db.max_batch_id() == 0
    assert db.next_batch_id() == 1
    assert db.max_batch_order(0) == 0
    assert db.next_batch_order(0) == 1

def test_artifact_row():
    a1 = Artifact()
    a1.key = "abc.txt"
    a1.batch_id = 1
    a1.batch_order = 1
    a1.source = 'run'
    a1.elapsed = 5
    a1.hashstring = "abcde123"

    db = SqliteDatabase(dbfile=None)
    db.append_artifact(a1)

    row = db.artifact_row(a1)
    assert row['key'] == a1.key
    assert row['batch_id'] == a1.batch_id

def test_append_and_update_artifacts():
    a1 = Artifact()
    a1.key = "abc.txt"
    a1.batch_id = 1
    a1.batch_order = 1
    a1.source = 'run'
    a1.elapsed = 5
    a1.hashstring = "abcde123"

    a2 = Artifact()
    a2.key = "def.txt"
    a2.hashstring = "abcde456"
    a2.batch_id = 1
    a2.batch_order = 2

    db = SqliteDatabase(dbfile=None)
    db.append_artifacts([a1, a2])

    assert len(db.conn.execute("select * from artifacts").fetchall()) == 2
    assert len(db.conn.execute("select * from artifacts where id = ?", (a1.unique_key(),)).fetchall()) == 1
    assert len(db.conn.execute("select * from artifacts where id = ?", (a2.unique_key(),)).fetchall()) == 1
    assert db.conn.execute("select elapsed from artifacts where id = ?", (a1.unique_key(),)).fetchall()[0][0] == 5

    a1.elapsed = 10

    db.update_artifact(a1)
    assert db.conn.execute("select elapsed from artifacts where id = ?", (a1.unique_key(),)).fetchall()[0][0] == 10



from dexy.dexy_filter import DexyFilter
from kaa import imlib2
import re

class Imlib2Filter(DexyFilter):
    """
    Base class which reads previous artifact, calls a do_im method, saves the
    output. Can be inherited. Default action is to flip an image, pointless in
    itself but a useful demo to know things are working.
    """
    ALIASES = ['im-flipv']
    BINARY = True
    FINAL = True
    FONT_PATHS = ["/usr/share/fonts/truetype/", "/usr/share/fonts/truetype/freefont/"]
    INPUT_EXTENSIONS = [".png", ".jpg", ".gif", ".tiff", ".bmp"]
    OUTPUT_EXTENSIONS = [".png", ".jpg", ".gif", ".tiff", ".bmp"]

    def do_im(self, image):
        image.flip_vertical()

    def process(self):
        font_paths = self.FONT_PATHS
        if self.artifact.args.has_key('font-path'):
            font_paths.extend(self.artifact.args['font-path'])

        for path in font_paths:
            imlib2.add_font_path(path)

        self.default_font = imlib2.load_font("FreeSans", 20)

        wf = self.artifact.previous_artifact_filepath
        image = imlib2.open(wf)
        new_image = self.do_im(image)

        # Most of the time, we can just modify the image in do_im, but in some
        # cases we need to return a new image object, i.e. when cropping. So,
        # if do_im returns anything replace the image with the returned image.
        if new_image:
            image = new_image

        image.save(self.artifact.filepath())

class Imlib2FlipHorizontalFilter(Imlib2Filter):
    ALIASES = ['im-fliph']

    def do_im(self, image):
        image.flip_horizontal()

class Imlib2GridFilter(Imlib2Filter):
    ALIASES = ['im-grid']

    def do_im(self, image):
        line_width = 3

        # draw vertical lines
        for i in xrange(100, image.width, 100):
            image.draw_rectangle((i, 0), (line_width, image.height), (0, 255, 0))
            image.draw_text((i+10, 0), "%s" % i, (255,0,0), self.default_font)

        # draw horizontal lines
        for j in xrange(100, image.height, 100):
            image.draw_rectangle((0, j), (image.width, line_width), (0, 0, 255))
            image.draw_text((0, j+10), "%s" % j, (255,0,0), self.default_font)

        for i in xrange(500, image.width, 500):
            for j in xrange(500, image.height, 500):
                image.draw_ellipse((i, j), (10, 10), (255,0,0))
                image.draw_text((i+15, j+15), "(%s,%s)" % (i, j), (255,0,0), self.default_font)

class Imlib2ScaleFilter(Imlib2Filter):
    ALIASES = ['im-scale']

    def do_im(self, image):
        if self.artifact.args.has_key('im-scale'):
            region_width, region_height = self.artifact.args['im-scale'].split("x")
        else:
            self.log.debug("Using default cropping region for im-scale, specify in args if you want something else")
            region_width=500
            region_height=500

        image = image.scale((int(region_width), int(region_height)))
        return image

class Imlib2Crop(Imlib2Filter):
    ALIASES = ['im-crop']

    def do_im(self, image):
        if self.artifact.args.has_key('im-crop'):
            crop_args = self.artifact.args['im-crop']
            regex = "([0-9]+)x([0-9]+)\+([0-9]+)\+([0-9]+)"
            if not re.match(regex, crop_args):
                raise Exception("Specify im-crop like 300x300+50+50.")

            region_size, start_x, start_y = crop_args.split("+")
            region_width, region_height = region_size.split("x")
            self.log.debug("Using custom cropping region for im-crop of width: %s height: %s x: %s y: %s" % (region_width, region_height, start_x, start_y))
        else:
            self.log.debug("Using default cropping region for im-crop, specify in args if you want something else")
            region_width=500
            region_height=500
            start_x=100
            start_y=100

        image = image.crop((int(start_x), int(start_y)), (int(region_width), int(region_height)))
        return image

class Imlib2Thumb(Imlib2Filter):
    ALIASES = ['im-thumb']

    def do_im(self, image):
        text_top = None
        width = None

        if self.artifact.args.has_key('im-thumb'):
            region_size, start_x, start_y = self.artifact.args['im-thumb']['crop'].split("+")
            region_width, region_height = region_size.split("x")
            text = self.artifact.args['im-thumb']['title']

            if self.artifact.args['im-thumb'].has_key('width'):
                width = self.artifact.args['im-thumb']['width']
            if self.artifact.args['im-thumb'].has_key('text-top'):
                text_top = self.artifact.args['im-thumb']['text-top']
        else:
            self.log.debug("Using default cropping region and text for im-thumb, specify in args if you want something else")
            region_width=500
            region_height=500
            start_x=100
            start_y=100
            text="Example"

        if not text_top:
            text_top = 10
        if not width:
            width=150 #final width

        image = image.crop((int(start_x), int(start_y)), (int(region_width), int(region_height)))
        image = image.scale((width, width))

        text_left = len(text)/2+30

        # Draw mock blank text to calculate height + width of text
        t_w, t_h, t_ha, t_va = image.draw_text((text_left, text_top), text, (255,255,255), self.default_font)

        # Draw white ellipse bigger than the ellipse people will see so it
        # doesn't run into neighboring text confusingly
        image.draw_ellipse((text_left+t_w/2, text_top+t_h/2), (t_w/2+25, t_h/2+20), (255,255,255), True)

        # Draw background ellipse
        image.draw_ellipse((text_left+t_w/2, text_top+t_h/2), (t_w/2+10, t_h/2+5), (238,221,130), True)

        # Draw actual text on top of ellipse
        image.draw_text((text_left, text_top), text, (0,0,0), self.default_font)

        return image




from boto.s3.key import Key
from dexy.dexy_filter import DexyFilter
import boto

class BotoUploadFilter(DexyFilter):
    """Uses boto library to upload content to S3."""
    ALIASES = ['botoup']

    # set your bucket name, env should have keys as per boto docs
    BUCKET_NAME = None

    def process_text(self, input_text):
        conn = boto.connect_s3()
        b = conn.get_bucket(self.BUCKET_NAME)
        k = Key(b)
        fn = self.artifact.filename(False)
        k.key = fn
        k.set_contents_from_filename(self.artifact.previous_artifact_filename)
        k.set_acl('public-read')
        return "https://s3.amazonaws.com/%s/%s" % (self.BUCKET_NAME, fn)



from dexy.artifact import Artifact
from dexy.artifacts.file_system_json_artifact import FileSystemJsonArtifact
from dexy.tests.utils import tempdir
from ordereddict import OrderedDict
import dexy.utils
import os

def is_empty_dict(d):
    return isinstance(d, dict) and not d

def test_empty_dict():
    x = {}
    assert is_empty_dict(x)
    x['a'] = 5
    assert not is_empty_dict(x)

def test_artifact_init():
    artifact = Artifact()
    assert artifact.state == 'new'

def test_artifact_filenames_simple_key():
    artifact = Artifact()
    artifact.key = 'abc'
    artifact.ext = '.out'
    assert artifact.canonical_filename() == 'abc.out'
    assert artifact.long_canonical_filename() == 'abc.out'

def test_artifact_filenames_file_key():
    artifact = Artifact()
    artifact.key = 'abc.txt'
    artifact.ext = '.out'
    assert artifact.canonical_filename() == 'abc.out'
    assert artifact.long_canonical_filename() == 'abc.txt.out'

def test_artifact_filenames_file_key_with_filters():
    artifact = Artifact()
    artifact.key = 'abc.txt|def|ghi'
    artifact.ext = '.out'
    assert artifact.canonical_filename() == 'abc.out'
    assert artifact.long_canonical_filename() == 'abc.txt-def-ghi.out'

def test_add_additional_artifact():
    db = dexy.utils.get_db(logsdir=None, dbfile=None)
    hashstring = "abcdef123"
    artifact = Artifact()
    artifact.db = db
    artifact.key = 'abc.txt'
    artifact.artifacts_dir = 'artifactsx'
    artifact.hashstring = hashstring

    new_artifact = artifact.add_additional_artifact('def.txt', '.txt')

    assert is_empty_dict(new_artifact._inputs)
    assert new_artifact.args.keys() == ["globals"]
    assert is_empty_dict(new_artifact.args['globals'])
    assert new_artifact.additional
    assert new_artifact.artifact_class_source
    assert new_artifact.artifacts_dir == 'artifactsx'
    assert new_artifact.final
    assert new_artifact.key in artifact.inputs()
    assert new_artifact.state == 'new'
    assert new_artifact.inode == hashstring

    assert db.artifact_row(new_artifact)['key'] == new_artifact.key

def test_simple_metadata():
    with tempdir():
        os.mkdir('artifacts')

        HASHSTRING = 'abc123'
        a1 = FileSystemJsonArtifact()
        a1.hashstring = HASHSTRING
        a1.key = 'xyz'
        a1.save_meta()

        a2 = FileSystemJsonArtifact()
        a2.hashstring = HASHSTRING
        assert not a2.key
        a2.load_meta()
        assert a2.key == 'xyz'

def test_artifact_data_dict_to_numbered_dict():
    artifact = Artifact()
    artifact.data_dict = OrderedDict()
    artifact.data_dict['a'] = 10
    artifact.data_dict['b'] = 20

    numbered_dict = artifact.convert_data_dict_to_numbered_dict()
    assert numbered_dict.keys()[0] == '00000:a'
    assert numbered_dict.keys()[1] == '00001:b'
    assert numbered_dict['00000:a'] == 10
    assert numbered_dict['00001:b'] == 20



from dexy.artifact import Artifact
from dexy.constants import Constants
from dexy.filters.templating_filters import JinjaFilterException
from dexy.filters.templating_filters import JinjaTextFilter

def init_jinja_filter():
    artifact = Artifact()
    artifact.name = "" # None causes problems with os.stat

    f = JinjaTextFilter()
    f.artifact = artifact
    f.log = Constants.NULL_LOGGER
    return f

def test_jinja_filter():
    f = init_jinja_filter()
    assert f.process_text("hello") == "hello"

def test_jinja_filter_with_basic_tag():
    f = init_jinja_filter()
    assert f.process_text("1 + 1 = {{ 1+1 }}") == "1 + 1 = 2"

def test_undefined_exception():
    f = init_jinja_filter()
    text = "{{ hello }}"
    f.artifact.input_data_dict = { '1' : text }
    f.artifact.previous_artifact_filepath = "file"
    try:
        f.process_text(text)
        assert False, "should not get here"
    except JinjaFilterException as e:
        print e.message

def test_undefined_exception_in_latex():
    f = init_jinja_filter()
    text = "<< hello >>"
    f.artifact.ext = ".tex"
    f.artifact.input_data_dict = { '1' : text }
    f.artifact.previous_artifact_filepath = "file"
    try:
        print f.process_text(text)
        assert False, "should not get here"
    except JinjaFilterException as e:
        print e.message

def test_syntax_error():
    text = "{{ [ }}"
    f = init_jinja_filter()
    f.artifact.input_data_dict = { '1' : text }
    f.artifact.previous_artifact_filepath = "file"

    try:
        f.process_text(text)
        assert False, "should not get here"
    except JinjaFilterException:
        pass

def test_syntax_error_in_latex():
    text = "<< [ >>"
    f = init_jinja_filter()
    f.artifact.ext = ".tex"
    f.artifact.input_data_dict = { '1' : text }
    f.artifact.previous_artifact_filepath = "file"

    try:
        f.process_text(text)
        assert False, "should not get here"
    except JinjaFilterException:
        pass



from dexy.artifact import Artifact
from dexy.dexy_filter import DexyFilter
from dexy.reporter import Reporter
import dexy.introspect
import inspect

def test_list_artifact_classes():
    artifact_classes = dexy.introspect.artifact_classes()
    assert len(artifact_classes) > 1
    for k, v in artifact_classes.iteritems():
        assert isinstance(k, str)
        assert issubclass(v, Artifact)

def test_list_filters():
    filters = dexy.introspect.filters()
    assert len(filters) > 1
    for k, v in filters.iteritems():
        assert isinstance(k, str)
        assert issubclass(v, DexyFilter)

def test_list_reporters():
    reporters = dexy.introspect.reporters()
    assert len(reporters) > 1
    for name, klass in reporters.iteritems():
        assert inspect.isclass(klass)
        assert issubclass(klass, Reporter)

def test_list_reports_dirs():
    reports_dirs = dexy.introspect.reports_dirs()
    assert len(reports_dirs) > 1
    for d in reports_dirs:
        assert isinstance(d, str)




from dexy.reporter import Reporter
import os

class ValidateReporter(Reporter):
    """
    A debugging reporter for checking that all elements in the
    database are also in the cache.
    """
    ALLREPORTS = False # TODO this breaks when repositories have multiple artifact types

    def run(self):
        for row in self.db.all():
            a = self.artifact_class.retrieve(row['hashstring'])
            if not os.path.exists(a.filepath()):
                raise Exception(a.key)
            else:
                print "%5d %s - ok" % (row['batch_id'], a.key)



from dexy.reporter import Reporter
from jinja2 import Environment
from jinja2 import FileSystemLoader
from ordereddict import OrderedDict
import cgi # for escape
import datetime
import os
import pstats
import random
import shutil
import sqlite3
import string
import web
import reporters.profile_reporter_files.aplotter as aplotter

try:
    import matplotlib
    matplotlib.use("Agg")
    import matplotlib.pyplot as pyplot
    MATPLOTLIB_AVAILABLE = True
except Exception:
    MATPLOTLIB_AVAILABLE = False

class ProfileReporter(Reporter):
    """
    Reporter which processes dexy.prof files generated by running dexy with
    cProfile. Builds up a history of profiling information each time it is run
    (info is saved in a sqlite database), so you can see changes over time.
    """
    ALLREPORTS = False
    DB_FILE = "profile.sqlite"
    REPORTS_DIR = 'logs/profile'
    DB_PATH = os.path.join(REPORTS_DIR, DB_FILE)

    def run(self):
        web.config.debug = False

        latest_dir = os.path.join(self.REPORTS_DIR, "profile-latest")
        shutil.rmtree(latest_dir, ignore_errors = True)

        if not os.path.exists(self.REPORTS_DIR):
            os.mkdir(self.REPORTS_DIR)

        db = web.database(dbn='sqlite', db=self.DB_PATH)
        try:
            db.query("""CREATE TABLE profiles(
                batchtimestamp text,
                filename text,
                lineno integer,
                functionname text,
                ncalls integer,
                primcalls integer,
                tottime real,
                cumtime real
            );""")
        except sqlite3.OperationalError:
            # table already exists, this is fine.
            pass

        ts = datetime.datetime.now().strftime("%Y-%m-%d--%H-%M-%S")
        report_dir = os.path.join(self.REPORTS_DIR, "profile-batch-%s-%s" % (self.batch_id, ts))
        os.mkdir(report_dir)

        if self.batch_info['args']['profile']: # This will raise an error if not run by profile_command since no default
            p = pstats.Stats('logs/dexy.prof')
            p.sort_stats('cumulative')

            env = Environment()
            env.loader = FileSystemLoader(os.path.dirname(__file__))
            template = env.get_template(os.path.join('profile_reporter_files', 'template.html'))

            function_data = OrderedDict()
            overall_tot_time = 0
            for i, x in enumerate(p.fcn_list):
                if i > 100:
                    break

                filename, lineno, functionname = x
                self.log.debug("processing profile data for item %s: function %s" % (i, functionname))
                ncalls, primcalls, tottime, cumtime, _ = p.stats[x]
                totpercall = tottime/ncalls
                cumpercall = cumtime/primcalls
                overall_tot_time += tottime

                # insert data from this run into db
                db.insert("profiles",
                    batchtimestamp=ts,
                    filename=filename,
                    lineno=lineno,
                    functionname=functionname,
                    ncalls=ncalls,
                    primcalls=primcalls,
                    tottime=tottime,
                    cumtime=cumtime
                    )
                short_filename = os.path.basename(filename)
                function_id = "%s:%s" % (cgi.escape(short_filename), cgi.escape(functionname))

                hist_rows = db.select("profiles",
                    where=("filename=\"%s\" AND functionname=\"%s\"" % (filename, functionname)),
                    order=("batchtimestamp ASC")
                )

                cumtime_hist = []
                tottime_hist = []
                for row in hist_rows:
                    cumtime_hist.append(row.cumtime)
                    tottime_hist.append(row.tottime)

                if MATPLOTLIB_AVAILABLE and i < 10:
                    pyplot.clf()
                    pyplot.plot(cumtime_hist)
                    cumtime_fig_filename = "%s.png" % ''.join(random.choice(string.ascii_uppercase + string.digits) for x in range(10))
                    figfile = open(os.path.join(report_dir, cumtime_fig_filename), "wb")
                    pyplot.savefig(figfile)
                    figfile.close()

                    pyplot.clf()
                    pyplot.plot(tottime_hist)
                    tottime_fig_filename = "%s.png" % ''.join(random.choice(string.ascii_uppercase + string.digits) for x in range(10))
                    figfile = open(os.path.join(report_dir, tottime_fig_filename), "wb")
                    pyplot.savefig(figfile)
                    figfile.close()
                else:
                    cumtime_fig_filename = None
                    tottime_fig_filename = None

                cumtime_text_plot = None
                tottime_text_plot = None
                if i < 20:
                    try:
                        cumtime_text_plot = aplotter.plot(cumtime_hist)
                        tottime_text_plot = aplotter.plot(tottime_hist)
                    except Exception:
                        pass

                function_data[function_id] = {
                    'functionname' : cgi.escape(functionname),
                    'ncalls' : ncalls,
                    'primcalls' : primcalls,
                    'filename' : filename,
                    'lineno' : lineno,
                    'tottime' : tottime,
                    'totpercall' : totpercall,
                    'cumtime' : cumtime,
                    'cumpercall' : cumpercall,
                    'cumtime_hist' : cumtime_hist,
                    'tottime_hist' : tottime_hist,
                    'cumtime_fig_filename' : cumtime_fig_filename,
                    'tottime_fig_filename' : tottime_fig_filename,
                    'cumtime_text_plot' : cumtime_text_plot,
                    'tottime_text_plot' : tottime_text_plot
                }

            env_data = {
                'function_data' : function_data,
                'overall_tot_time' : overall_tot_time
            }
            template.stream(env_data).dump(os.path.join(report_dir, 'index.html'))
            shutil.copytree(report_dir, latest_dir)




from dexy.dexy_filter import DexyFilter
from ordereddict import OrderedDict
from pygments import highlight
from pygments.formatters import get_formatter_for_filename
from pygments.lexers import get_lexer_by_name
from pygments.lexers import get_lexer_for_filename
from pygments.lexers.agile import PythonConsoleLexer
from pygments.lexers.agile import RubyConsoleLexer
from pygments.lexers.text import TexLexer
from pygments.lexers.web import JavascriptLexer
from pygments.lexers.web import XmlLexer
import pygments.lexers.web

class PygmentsFilter(DexyFilter):
    """
    Apply Pygments syntax highlighting. Image formats require PIL.
    """
    INPUT_EXTENSIONS = [".*"]
    IMAGE_OUTPUT_EXTENSIONS = ['.png', '.bmp', '.gif', '.jpg']
    MARKUP_OUTPUT_EXTENSIONS = [".html", ".tex", ".svg"]
    OUTPUT_EXTENSIONS = MARKUP_OUTPUT_EXTENSIONS + IMAGE_OUTPUT_EXTENSIONS
    ALIASES = ['pyg', 'pygments']
    FINAL = False

    def process_dict(self, input_dict):
        ext = self.artifact.input_ext
        has_args = self.artifact.args.has_key('pygments')
        if has_args:
            args = self.artifact.args['pygments']
        else:
            args = {}

        if args.has_key('lexer'):
            lexer = get_lexer_by_name(args['lexer'])
            del args['lexer']
        else:
            if ext == '.pycon':
                lexer = PythonConsoleLexer()
            elif ext == '.rbcon':
                lexer = RubyConsoleLexer()
            elif (ext in ('.json', '.dexy') or self.artifact.name.endswith(".dexy")) and (pygments.__version__ < '1.5'):
                lexer = JavascriptLexer()
            elif ext in ('.dexy') or self.artifact.name.endswith(".dexy"):
                # JSON lexer available in pygments 1.5
                lexer = pygments.lexers.web.JSONLexer()
            elif ext == '.Rd':
                lexer = TexLexer() # does a passable job
            elif ext == '.svg':
                lexer = XmlLexer()
            else:
                fake_file_name = "input_text%s" % ext
                lexer = get_lexer_for_filename(fake_file_name)

        formatter_args = {'lineanchors' : self.artifact.web_safe_document_key() }
        # all args left are for the formatter...
        formatter_args.update(args)

        formatter = get_formatter_for_filename(self.artifact.filename(),
                **formatter_args)
        ### @end
        # TODO whitelist acceptable formatter args
        # TODO allow passing lexer args?

        if self.artifact.ext in self.IMAGE_OUTPUT_EXTENSIONS:
            self.artifact.binary_output = True
            # TODO set to final
            with open(self.artifact.filepath(), 'wb') as f:
                f.write(highlight(self.artifact.input_text(), lexer, formatter))
        else:
            output_dict = OrderedDict()
            for k, v in input_dict.items():
                # TODO figure out where these characters are coming from and don't hard-code this.
                v = v.replace(" \x08", "").replace(chr(13), "")
                output_dict[k] = highlight(v, lexer, formatter)
            return output_dict



import logging

try:
    from logging import NullHandler
except:
    # NullHandler not in Python 2.6
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

class Constants():

    # Create a null logger for testing and other times when we don't want to
    # worry about logfiles
    NULL_LOGGER = logging.getLogger("null")
    NULL_LOGGER.addHandler(NullHandler())
    NULL_LOGGER.propagate = 0

    LOGLEVELS = {
        'DEBUG' : logging.DEBUG,
        'INFO' : logging.INFO,
        'WARNING' : logging.WARNING,
        'ERROR' : logging.ERROR,
        'CRITICAL' : logging.CRITICAL
    }

    DEFAULT_ACLASS = 'FileSystemJsonArtifact'
    DEFAULT_ADIR = 'artifacts'
    DEFAULT_COMMAND = 'dexy'
    DEFAULT_CONFIG = '.dexy'
    DEFAULT_LDIR = 'logs'
    DEFAULT_REPORTS = "Output LongOutput Run Source"
    DEFAULT_LOGGER_NAME = 'dexy'
    DEFAULT_LFILE = 'dexy.log'
    DEFAULT_DBCLASS = 'SqliteDatabase'
    DEFAULT_DBFILE = "db.sql"
    DEFAULT_LOGLEVEL="DEBUG"
    DEFAULT_LOGFORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

    # Directories with this name should be excluded anywhere
    EXCLUDE_DIRS_ALL_LEVELS = ['.bzr', '.hg', '.git', '.svn']

    # Directories with these names should be excluded only at the project root
    EXCLUDE_DIRS_ROOT = ['ignore']

    EXCLUDE_HELP = """Specify directory names to exclude from processing by dexy.
    Directories with these names will be skipped anywhere in your project.
    The following patterns are automatically excluded anywhere in your project: %s.
    The directories designated for artifacts and logs are automatically excluded,
    and the following directory names are also excluded if they appear at the root
    level of your project: %s.
    Any directory with a file named .nodexy will be skipped, and subdirectories of this will also be skipped.
    """ % (", ".join(EXCLUDE_DIRS_ALL_LEVELS), ", ".join(EXCLUDE_DIRS_ROOT))

    ARTIFACT_HASH_WHITELIST = [
        'args',
        'artifact_class_source',
        'ctime',
        'dexy_version',
        'dirty',
        'dirty_string',
        'ext',
        'filter_name',
        'filter_source',
        'filter_version',
        'inode',
        'input_data_dict',
        'input_ext',
        'inputs',
        'key',
        'mtime',
        'next_filter_name'
    ]



from dexy.artifacts.file_system_json_artifact import FileSystemJsonArtifact
import pickle

class FileSystemPickleArtifact(FileSystemJsonArtifact):
    def meta_filename(self):
        return "%s-meta.pickle" % (self.hashstring)

    def cached_output_filename(self):
        return "%s-output.pickle" % (self.hashstring)

    def write_dict_to_file(self, data_dict, filepath):
        with open(filepath, "wb") as f:
            pickle.dump(data_dict, f)

    def read_dict_from_file(self, filepath):
        with open(filepath, "rb") as f:
            return pickle.load(f)




from dexy.constants import Constants
from dexy.reporter import Reporter
from dexy.tests.utils import tempdir
from dexy.utils import remove_all_handlers
from dexy.utils import get_log
from logging import StreamHandler
from logging.handlers import RotatingFileHandler
import os

def setup():
    log = get_log(logsdir=None, logfile=None)
    remove_all_handlers(log)

def test_logging_to_stderr():
    r = Reporter(None, None, dbfile=None)
    assert len(r.log.handlers) == 1
    assert isinstance(r.log.handlers[0], StreamHandler)
    assert not isinstance(r.log.handlers[0], RotatingFileHandler)
    remove_all_handlers(r.log)

def test_logging_to_file():
    with tempdir():
        os.mkdir(Constants.DEFAULT_LDIR)
        r = Reporter()
        assert len(r.log.handlers) == 1
        assert isinstance(r.log.handlers[0], RotatingFileHandler)
        remove_all_handlers(r.log)



from dexy.reporter import Reporter
from jinja2 import Environment
from jinja2 import FileSystemLoader
from pygments import highlight
from pygments.formatters.html import HtmlFormatter
from pygments.lexers.web import JavascriptLexer
import datetime
import json
import os
import shutil

class RunReporter(Reporter):
    def run(self):
        self.load_batch_artifacts()

        # Create a directory to hold the report.
        report_dir = os.path.join(self.logsdir, "run-%s" %
                                  datetime.datetime.now().strftime("%Y-%m-%d--%H-%M-%S"))
        latest_report_dir = os.path.join(self.logsdir, "run-latest")
        report_filename = os.path.join(report_dir, 'index.html')
        shutil.rmtree(report_dir, ignore_errors=True)

        template_dir = os.path.join(os.path.dirname(__file__), 'run_reporter')
        shutil.copytree(template_dir, report_dir)

        formatter = HtmlFormatter()
        js_lexer = JavascriptLexer()

        env_data = {}

        def highlight_json(d):
            j = json.dumps(d, sort_keys=True, indent=4)
            return highlight(j, js_lexer, formatter)

        total_artifacts = 0
        base_artifacts = 0
        cached_artifacts = 0
        run_artifacts = 0

        total_elapsed = 0
        total_run_elapsed = 0
        run_time_by_key = {}
        for key, doc in self.batch_info['docs'].iteritems():
            for hashstring, source, elapsed in doc['artifacts']:

                artifact = self.artifacts[hashstring]

                if hasattr(artifact, 'filter_name'):
                    filter_class = artifact.filter_name
                    if not run_time_by_key.has_key(filter_class):
                        run_time_by_key[filter_class] = []

                    elapsed_array = run_time_by_key[filter_class]
                    elapsed_array.append(elapsed)
                    run_time_by_key[filter_class] = elapsed_array

                total_artifacts += 1
                total_elapsed += elapsed

                if not source:
                    base_artifacts += 1
                elif source == 'run':
                    run_artifacts += 1
                    total_run_elapsed += elapsed
                elif source =='cache':
                    cached_artifacts += 1
                else:
                    raise Exception("unknown source %s" % source)

        env_data['total_artifacts'] = total_artifacts
        env_data['base_artifacts'] = base_artifacts
        env_data['cached_artifacts'] = cached_artifacts
        env_data['run_artifacts'] = run_artifacts
        env_data['total_elapsed'] = total_elapsed
        env_data['total_run_elapsed'] = total_run_elapsed
        env_data['run_time_by_key'] = run_time_by_key

        env_data['artifacts'] = self.artifacts
        env_data['batch_id'] = self.batch_id
        env_data['batch_info'] = self.batch_info
        env_data['dexy_config'] = highlight_json(self.batch_info['config'])
        env_data['docs'] = self.batch_info['docs']
        env_data['highlight_json'] = highlight_json

        env_data['sorted'] = sorted
        env_data['float'] = float
        env_data['min'] = min
        env_data['max'] = max
        env_data['len'] = len
        env_data['sum'] = sum

        env = Environment()
        env.loader = FileSystemLoader(os.path.dirname(__file__))
        template = env.get_template('run_reporter_template.html')
        template.stream(env_data).dump(report_filename)

        shutil.rmtree(latest_report_dir, ignore_errors=True)
        shutil.copytree(report_dir, latest_report_dir)



import dexy.filters.blog_filters

# TODO write tumblr access from scratch as e.g. saving as draft raises an error
# http://code.google.com/p/python-tumblr/
from tumblr import Api
class TumblrFilter(dexy.filters.blog_filters.BlogFilter):
    """
    Posts to a tumblr blog.
    """
    ALIASES = ['tumblr']
    BLOG_CONFIG_FILE = 'tumblr-config.json'

    def initialize_api(self):
        blog = self.blog_conf['blog']
        user = self.blog_conf['user']
        password = self.blog_conf['password']
        return Api(blog, user, password)

    def new_post(self, api, input_text):
        title = self.post_conf.pop('title')
        post = api.write_regular(title, input_text, **self.post_conf)
        return post['post-id']

    def update_post(self, api, input_text, post_id):
        raise Exception("not implemented!")




from dexy.dexy_filter import DexyFilter
import os
import subprocess

class ProcessFilter(DexyFilter):
    """
    Base class for all classes that do external processing, via subprocess or pexpect.
    """
    ALIASES = ['processfilter']

    def handle_subprocess_proc_return(self, returncode, stderr):
        if returncode is None:
            raise Exception("no return code, proc not finished!")
        elif returncode != 0:
            artifact_ignore = self.artifact.args.has_key('ignore') and self.artifact.args['ignore']
            controller_ignore = self.artifact.controller_args['ignore']
            if self.IGNORE_ERRORS or artifact_ignore or controller_ignore:
                self.artifact.log.warn("Nonzero exit status %s" % returncode)
                self.artifact.log.warn("output from process: %s" % stderr)
            else:
                print stderr
                raise Exception("""proc returned nonzero status code! if you don't want dexy to raise errors on failed scripts then pass the -ignore option""")

    def setup_env(self):
        if self.artifact.args.has_key('env'):
            env = os.environ
            env.update(self.artifact.args['env'])
        else:
            env = None
        return env

    def setup_timeout(self):
        if self.artifact.args.has_key('timeout'):
            timeout = self.artifact.args['timeout']
            self.log.info("using custom timeout %s for %s" % (timeout, self.artifact.key))
        else:
            timeout = None
        return timeout

    def setup_cwd(self):
        cwd = self.artifact.artifacts_dir
        if self.artifact.args.has_key('cwd'):
            cwd = os.path.join(cwd, self.artifact.args['cwd'])
            self.log.debug("Changing into directory %s" % os.path.abspath(cwd))
        return cwd

    def command_line_args(self):
        """
        Allow specifying command line arguments which are passed to the filter
        with the given key. Note that this does not currently allow
        differentiating between 2 calls to the same filter in a single document.
        """
        if self.artifact.args.has_key('args'):
            args = self.artifact.args['args']
            last_key = self.artifact.key.rpartition("|")[-1]
            if args.has_key(last_key):
                return args[last_key]

    def command_line_scriptargs(self):
        """
        Allow specifying command line arguments which are passed to the filter
        with the given key, to be passed as arguments to the script being run
        rather than the executable.
        """
        if self.artifact.args.has_key('scriptargs'):
            args = self.artifact.args['scriptargs']
            last_key = self.artifact.key.rpartition("|")[-1]
            if args.has_key(last_key):
                return args[last_key]

    def command_string(self):
        args = {
            'prog' : self.executable(),
            'args' : self.command_line_args() or "",
            'scriptargs' : self.command_line_scriptargs() or "",
            'script_file' : self.artifact.previous_artifact_filename,
            'output_file' : self.artifact.filename()
        }
        return "%(prog)s %(args)s %(script_file)s %(scriptargs)s %(output_file)s" % args

    def command_string_stdout(self):
        args = {
            'prog' : self.executable(),
            'args' : self.command_line_args() or "",
            'scriptargs' : self.command_line_scriptargs() or "",
            'script_file' : self.artifact.previous_artifact_filename
        }
        return "%(prog)s %(args)s %(script_file)s %(scriptargs)s" % args

class SubprocessFilter(ProcessFilter):
    ALIASES = ['subprocessfilter']
    BINARY = True
    FINAL = True

    def process(self):
        command = self.command_string()
        proc, stdout = self.run_command(command, self.setup_env())
        self.handle_subprocess_proc_return(proc.returncode, stdout)
        self.artifact.stdout = stdout

    def run_command(self, command, env, input_text = None):
        cwd = self.setup_cwd()
        self.log.debug("about to run '%s' in %s" % (command, cwd))
        proc = subprocess.Popen(command, shell=True,
                                cwd=cwd,
                                stdin=subprocess.PIPE,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                env=env)

        stdout, stderr = proc.communicate(input_text)
        return (proc, stdout)

class SubprocessStdoutFilter(SubprocessFilter):
    ALIASES = ['subprocessstdoutfilter']
    BINARY = False
    FINAL = False

    def process(self):
        command = self.command_string_stdout()
        self.log.debug("About to run '%s'" % command)
        proc, stdout = self.run_command(command, self.setup_env())
        self.handle_subprocess_proc_return(proc.returncode, stdout)
        self.artifact.set_data(stdout)

class SubprocessStdoutInputFilter(SubprocessFilter):
    ALIASES = ['subprocessstdoutinputfilter']
    BINARY = False
    FINAL = False

    def process(self):
        command = self.command_string_stdout()

        if len(self.artifact.inputs()) == 1:
            artifact = self.artifact.inputs().values()[0]
            for section_name, section_text in artifact.data_dict.iteritems():
                proc, stdout = self.run_command(command, self.setup_env(), section_text)
                self.handle_subprocess_proc_return(proc.returncode, stdout)
                self.artifact.data_dict[section_name] = stdout
        else:
            for artifact in self.artifact.inputs().values():
                proc, stdout = self.run_command(command, self.setup_env(), artifact.output_text())
                self.handle_subprocess_proc_return(proc.returncode, stdout)
                rel_key = self.artifact.relative_key_for_input(artifact)
                self.artifact.data_dict[rel_key] = stdout

class SubprocessStdoutInputFileFilter(SubprocessFilter):
    ALIASES = ['subprocessstdoutinputfilefilter']
    BINARY = False
    FINAL = False

    def command_string_stdout_input(self, input_artifact):
        script_file = self.artifact.previous_artifact_filename
        input_file = input_artifact.filename()
        args = self.command_line_args() or ""
        return "%s %s %s %s" % (self.executable(), args, script_file, input_file)

    def process(self):
        for artifact in self.artifact.inputs().values():
            command = self.command_string_stdout_input(artifact)
            proc, stdout = self.run_command(command, self.setup_env())
            self.handle_subprocess_proc_return(proc.returncode, stdout)
            rel_key = self.artifact.relative_key_for_input(artifact)
            self.artifact.data_dict[rel_key] = stdout

class SubprocessCompileFilter(SubprocessFilter):
    """
    Base class for filters which need to compile code, then run the compiled executable.
    """
    ALIASES = ['subprocesscompilefilter']
    BINARY = False
    FINAL = False
    COMPILED_EXTENSION = ".o"
    CHECK_RETURN_CODE = False # Whether to check return code when running compiled executable.

    def compile_command_string(self):
        wf = self.artifact.previous_artifact_filename
        of = self.artifact.temp_filename(self.COMPILED_EXTENSION)
        return "%s %s -o %s" % (self.executable(), wf, of)

    def run_command_string(self):
        of = self.artifact.temp_filename(self.COMPILED_EXTENSION)
        return "./%s" % of

    def process(self):
        # Compile the code
        command = self.compile_command_string()
        proc, stdout = self.run_command(command, self.setup_env())
        self.handle_subprocess_proc_return(proc.returncode, stdout)

        command = self.run_command_string()
        proc, stdout = self.run_command(command, self.setup_env())
        if self.CHECK_RETURN_CODE:
            self.handle_subprocess_proc_return(proc.returncode, stdout)

        self.artifact.set_data(stdout)

class SubprocessCompileInputFilter(SubprocessCompileFilter):
    ALIASES = ['subprocesscompileinputfilter']
    CHECK_RETURN_CODE = False

    def process(self):
        # Compile the code
        command = self.compile_command_string()
        proc, stdout = self.run_command(command, self.setup_env())
        self.handle_subprocess_proc_return(proc.returncode, stdout)

        command = self.run_command_string()

        if len(self.artifact.inputs()) == 1:
            artifact = self.artifact.inputs().values()[0]
            for section_name, section_text in artifact.data_dict.iteritems():
                proc, stdout = self.run_command(command, self.setup_env(), section_text)
                if self.CHECK_RETURN_CODE:
                    self.handle_subprocess_proc_return(proc.returncode, stdout)
                self.artifact.data_dict[section_name] = stdout
        else:
            for artifact in self.artifact.inputs().values():
                proc, stdout = self.run_command(command, self.setup_env(), artifact.output_text())
                if self.CHECK_RETURN_CODE:
                    self.handle_subprocess_proc_return(proc.returncode, stdout)
                rel_key = self.artifact.relative_key_for_input(artifact)
                self.artifact.data_dict[rel_key] = stdout



import mako.template
import dexy.filters.templating_filters

class MakoFilter(dexy.filters.templating_filters.TemplateFilter):
    ALIASES = ['mako']

    def process_text(self, input_text):
        return str(mako.template.Template(input_text).render(self.run_plugins()))
